/*
 * mid2seq - Covert a MIDI file to a Zoom SEQ file.
 *
 *
 * Copyright (C) 2003-2004 Dwight Engen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <libzoom.h>
#include "midifile.h"

#ifndef MIN
#define MIN(x,y) ((x) > (y) ? (y) : (x))
#endif
#ifndef MAX
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#endif


#define MAX_SMF_MEASURES	2048
#define MEASURES_PER_PATTERN	1

int		 verbose = 0;
ps04_seq_t	*seq;
int		 seq_ppq = 0;
int		 ptn_ppq_last;
int		 ptn_max;
int		 ptn_dups = 0;

FILE		*SMF;
int		 smf_timesig_numer;
int		 smf_ppq;
double		 ppq_scale;

ps04_pattern_t	*flat_song[MAX_SMF_MEASURES];

static void usage(char *prgname)
{
    printf("mid2seq v1.00 Copyright (C) 2003-2004 Dwight Engen\n\n");
    printf("This program is free software; you can redistribute it and/or modify\n"
	   "it under the terms of the GNU General Public License as published by\n"
	   "the Free Software Foundation; either version 2 of the License, or\n"
	   "(at your option) any later version.\n\n");

    printf("Convert a MIDI SMF to a PS04 RhythmXX.seq\n");
    printf("Usage: %s [OPTION] <file.mid> <output.seq>\n\n", prgname);
    printf(" -h  --help\n");
    printf(" -v  --verbose\n\n");

    exit(EXIT_FAILURE);
}



#if 0
static const char *PS04_event_name(song_event_type_t event)
{
    static const char event_name[][9] =
    {
	"Invalid",
	"Pattern",
	"Note",
	"Tempo",
	"Timesig",
	"Drum Vol",
	"Bass Vol",
	"Drum Kit",
	"Bass Kit",
    };
    if (event < 9)
	return event_name[event];
    else
	return "Unknown";
}



static void PS04_song_dump(ps04_song_t *song, ps04_seq_t  *seq)
{
    int i;
    int ppq;

    printf("Song    : %-8.8s\n", song->name);
    printf("Events  : %d\n", song->events);

    ppq = 0;
    for (i = 0; i < song->events; i++)
    {
	unsigned short data = song->event[i].data;

	printf("Event:%4.4d PPQ:%4.4d (%5.5d) Meas:%3.3d Type:%-10s ",
		i+1, song->event[i].next_ppq, ppq,
		ppq/(PS04_PPQ*smf_timesig_numer)+1,
		PS04_event_name(song->event[i].type));

	switch(song->event[i].type)
	{
	case SONG_EVENT_PATTERN:
	    printf("%8.8s (0x%3.3x)\n", seq->pattern[data]->name, data);
	    break;
	case SONG_EVENT_NOTE:
	    printf("%-2s%3s\n", note_root_name[data & 0x0f],
		   note_tone_name[(data & 0xf0) >> 4]);
	    break;

	case SONG_EVENT_TEMPO:
	    printf("%d.%d\n", data/10, data - ((data/10)*10));
	    break;

	case SONG_EVENT_TIMESIG:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_DRUM_VOL:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_BASS_VOL:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_DRUM_KIT:
	    break;
	case SONG_EVENT_BASS_KIT:
	    break;
	}

	/* Handle large/small PPQ's */
	if (song->event[i].next_ppq == 0xff)
	    ppq += *(uint32_t *)&song->event[++i];
	else
	    ppq += song->event[i].next_ppq;
    }
    printf("\n");
}
#endif



static int SMFError(const char *msg)
{
    fprintf(stderr, "SMF error %s\n", msg);
    return 0;
}

static int SMFGetC(void)
{
    return getc(SMF);
}

static int SMFHeader(int format, int tracks, int division)
{
    smf_ppq = division;
    if (smf_ppq > PS04_PPQ)
    {
	ppq_scale = (double)smf_ppq / (double)PS04_PPQ;
	if ((smf_ppq % PS04_PPQ) != 0)
	    printf("Warning: SMF PPQ %d not multiple of Zoom hardware PPQ %d scale:%f\n",
		   smf_ppq, PS04_PPQ, ppq_scale);
    }

    printf("SMFFormat: %d\n", format);
    printf("SMFTracks: %d\n", tracks);
    printf("SMFPPQ:    %d\n", smf_ppq);
    return 0;
}

static ps04_pattern_t *ptn_alloc(int measure)
{
    ps04_pattern_t	*ptn;
    int			 ptn_indx;

    Assert(measure >= 0 && measure < MAX_SMF_MEASURES, "Out of bounds measure %d\n", measure);

    ptn_indx = measure / MEASURES_PER_PATTERN;
    ptn_max = MAX(ptn_max, ptn_indx);
    ptn = flat_song[ptn_indx];

    if (!strcmp(ptn->name_long, "(empty)"))
    {
	sprintf(ptn->name,      "Flat%4.4d", ptn_indx);
	sprintf(ptn->name_long, "Flat%4.4d", ptn_indx);

	ptn->beats = smf_timesig_numer;
	ptn->bars = MEASURES_PER_PATTERN;
	ptn->drum_kit = 0;
	ptn->bass_kit = 0;
	ptn->drum_level = 0x0f;
	ptn->bass_level = 0x0f;
	ptn->starting_ppq = measure * PS04_PPQ * smf_timesig_numer;
	ptn_ppq_last = 0;
    }
    return ptn;
}

static int MIDI_Note_2_PS04(int note)
{
    switch(note)
    {
    case 35:	return 36;	/* Kick 2     -> Kick 1 */
    case 40:	return 38;	/* Snare 2    -> Snare 1 */
    case 44:	return 46;	/* Pedal Hat  -> Open Hat */
    case 52:	return 53;	/* China Cym  -> Ride bell (Ug) */
    case 55:	return 49;	/* Splash Cym -> Crash 1 (Ug) */
    case 57:	return 49;	/* Crash 2    -> Crash 1 */
    case 59:	return 51;	/* Ride 2     -> Ride 1 */
    default:
	return note;
    }
}


static void emit_ptn_event(ps04_pattern_t *ptn, int abs_ppq, int type, int note, int vel)
{
    ptn_event_t *ev;
    int ptn_ppq;
    
    ptn->event = realloc(ptn->event, (ptn->sevents+1) * sizeof(*ptn->event));
    ev = &ptn->event[ptn->sevents++];

    ev->type = type;
    ptn_ppq = abs_ppq - ptn->starting_ppq;

    ev->ppq_delta = ptn_ppq - ptn_ppq_last;
    ptn_ppq_last = ptn_ppq;
    ev->ppq_duration = 48;
    ev->midi_note = note == 0 ? 0 : MIDI_Note_2_PS04(note);
    ev->velocity15 = vel >> 3;
}



static int SMFNoteOn(int chan, int note, int vel)
{
    int ppq;
    int measure;
    static int last_measure = -1;
    ps04_pattern_t *ptn;

    if (chan != 9)
	return 0;
    if (vel == 0)
	return 0;

    measure = (Mf_currtime / smf_ppq) / smf_timesig_numer;

    /* If we've entered a new measure, fixup previous measures remaining
     * delta.
     */
    if (measure != last_measure && measure != 0)
    {
	ptn = ptn_alloc(measure-1);
	ppq = measure * smf_timesig_numer * PS04_PPQ;
	emit_ptn_event(ptn, ppq, PTN_EVENT_END, 0x0f, 0);
	last_measure = measure;
    }

    ppq = Mf_currtime / ppq_scale;
    ptn = ptn_alloc(measure);
    emit_ptn_event(ptn, ppq, PTN_EVENT_DRUM, note, vel);

#if 0
    printf("SMFTick: %6.6d PPQ:%4.4d Meas:%3.3d Vel:%-3d Note:%d %s\n",
	   Mf_currtime, ppq, measure+1, vel, note, PS0x_midi_drum_name(note));
#endif
    return 0;
}

static int SMFTimeSig(int numer, int denom_log2, int cc, int bb)
{
    int denom = 1;
    while ( denom_log2-- > 0 )
	    denom *= 2;
    printf("Time signature=%d/%d  MIDI-clocks/click=%d  32nd-notes/24-MIDI-clocks=%d\n",
	    numer, denom, cc, bb);
    smf_timesig_numer = numer;
    return 0;
}


static void MFuncsInit(void)
{
    Mf_getc = SMFGetC;
    Mf_error = SMFError;
    Mf_header = SMFHeader;
    Mf_timesig =  SMFTimeSig;
    Mf_noteon =  SMFNoteOn;
}



/* Determine if the given pattern p2 is a duplicate of any pattern
 * already put into the SEQ file, if it is return the index of the
 * SEQ file pattern, otherwise return -1.
 */
static int ptn_in_seq(ps04_pattern_t *p2)
{
    int i;
    ps04_pattern_t *p1;

    for(i = 0; i < seq->patterns; i++)
    {
	p1 = seq->pattern[i];

	if (p1->event != NULL &&
	    p1->sevents == p2->sevents &&
	    !memcmp(p1->event, p2->event, p1->sevents * sizeof(*p1->event)))
	{
	    ptn_dups++;
	    return i;
	}
    }
    return -1;
}

static void emit_song_event(ps04_song_t *song, int type, int data, int next_ppq)
{
    song_event_t *ev;
    
    song->event = realloc(song->event, (song->events+1) * sizeof(*song->event));
    ev = &song->event[song->events++];

    ev->type = type;
    ev->data = data;
    ev->next_ppq = next_ppq;
}


int main(int argc, char *argv[])
{
    int opt;
    int seq_ptns;
    int i;

    static struct option long_options[] =
    {
	{"help",		0, 0, '?'},
	{"verbose",		0, 0, 'v'},

	{0,0,0,0},
    };

    if (argc < 3)
	usage(argv[0]);

    while ((opt = getopt_long(argc, argv, "v", long_options, NULL)) != -1)
    {
	switch(opt)
	{
	default:
	case '?':
	    usage(argv[0]);
	    break;
	case 'v':
	    verbose = 1;
	    break;
	}
    }

    if (argc - optind != 2)
	usage(argv[0]);

    SMF = fopen(argv[optind], "rb");
    if (SMF == NULL)
    {
	fprintf(stderr, "Unable to open %s\n", argv[optind]);
	exit(EXIT_FAILURE);
    }
    optind++;

    seq = PS04_seq_new();
    seq->F = fopen(argv[optind], "w+b");
    if (seq->F == NULL)
    {
	fprintf(stderr, "Unable to open %s\n", argv[optind]);
	exit(EXIT_FAILURE);
    }

    /* Initialize the flat song */
    for (i = 0; i < MAX_SMF_MEASURES; i++)
    {
	flat_song[i] = PS04_pattern_new();
    }

    MFuncsInit();
    mfread();

    /* Now we have all the patterns made into a flat song, lets make
     * the SEQ song
     */

    if (verbose)
	printf("Converted song into %d flat patterns (measures)\n", ptn_max);

    strcpy(seq->song[0]->name_long, "mid2seq");

    /* FIXME: emit_song_event(seq->song[0], SONG_EVENT_TEMPO, ); */
    emit_song_event(seq->song[0], SONG_EVENT_TIMESIG, smf_timesig_numer, 0);
    seq_ptns = 0;
    for (i = 0; i < ptn_max; i++)
    {
	int seq_ptn;

	/* See if this pattern is an exact duplicate of one we
	 * already used, if so use that one.
	 */
	seq_ptn = ptn_in_seq(flat_song[i]);
	if (seq_ptn < 0)
	{
	    ps04_pattern_t *ptn;

	    Assert(seq_ptns <= 511, "Too many patterns for SEQ file!\n");

	    /* Add this pattern to the SEQ file */
	    seq_ptn = seq_ptns;
	    ptn = flat_song[i];
	    sprintf(ptn->name,      "Ptn%3.3d", seq_ptn);
	    sprintf(ptn->name_long, "Ptn%3.3d", seq_ptn);
	    seq->pattern[seq_ptns++] = flat_song[i];
	}

	/* Output song event for this pattern */
	emit_song_event(seq->song[0], SONG_EVENT_PATTERN, seq_ptn,
			PS04_PPQ * smf_timesig_numer * MEASURES_PER_PATTERN);
    }

    if (verbose)
    {
	printf("SMF had %d duplicate drum patterns\n", ptn_dups);
	printf("SEQ has %d unique drum patterns\n", seq_ptns);
    }


#if 0
    if (verbose)
    {
	ps02_song_t *song02;

	song02 = pts->song[songnr-1];
	printf("Song    : %-6.6s\n", song02->name);
	printf("Tempo   : %d\n", song02->tempo);

	for(i =  0; i < pts->rhy_pattern_cnt; i++)
	    printf("Pattern : %6.6s\n", pts->rhy_pattern[i]->name);

	for(i = 0; i < song02->measures; i++)
	    printf("Measure : %3.3d pattern:%6.6s\n", i+1,
		   pts->rhy_pattern[song02->measure[i].pattern]->name);
    }

    PS02_to_PS04(pts, seq, songnr-1);
    //PS04_song_dump(seq->song[0], seq);
#endif

    //PS04_song_dump(seq->song[0], seq);
    PS04_seq_write(seq);
    PS04_seq_close(seq);
    return EXIT_SUCCESS;
}
