/*
 * mrs8sample
 *
 * Copyright (C) 2003-2005 Dwight Engen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <libzoom.h>

#define CMD_DUMP_INFO		0x000000001
#define CMD_EXPORT_SAMPLES	0x000000002
#define CMD_IMPORT_SAMPLES	0x000000004

static mrs8_bin_t *bin;

static void usage(char *prgname)
{
    printf("mrs8sample v1.00 Copyright (C) 2003-2005 Dwight Engen\n\n");
    printf("This program is free software; you can redistribute it and/or modify\n"
	   "it under the terms of the GNU General Public License as published by\n"
	   "the Free Software Foundation; either version 2 of the License, or\n"
	   "(at your option) any later version.\n\n");

    printf("Usage: %s <OPTION> <file.bin>\n\n", prgname);
    printf(" -h  --help\n\n");

    printf(" -d  --dump-info\n");
    printf(" -x  --export-samples\n");
    printf(" -i  --import-samples\n");

    exit(EXIT_FAILURE);
}

static int32_t MRS8_chksum(int8_t byte, int reset)
{
    static int32_t chksum;

    if (reset)
	chksum = 0;
    chksum += byte;
    return chksum;
}

static void MRS8_dump_info(mrs8_bin_t *bin)
{
    int		 i;
    int		 j;
    int		 frame_cnt;

    for(i = 0; i < 7; i++)
    {
	mrs8_drum_kit_t *kit = &bin->drum_kits.kit[i];
	printf("Kit Name: %8.8s\n", kit->name);

	printf("%-5s %-7s %-4s %-5s %-6s %s\n",
	       "Slot", "Sample", "Pan", "Vol?", "Pitch", "MIDI");
	for(j = 0; j < MRS8_DRUMS_PER_KIT; j++)
	{
	    mrs8_drum_t *drum = &kit->drum[j];

	    if (drum->sample == 0x38)
		continue;

	    printf("%-5d %-7d %-4d %-5d %-6d %2.2d %s\n",
		   j, drum->sample, drum->pan, drum->unknown1,
		   drum->pitch, drum->midi_note,
		   PS0x_midi_drum_name(drum->midi_note));
	}
	printf("\n");
    }

    printf("Samples: %d\n", bin->samples);
    printf("%-7s %-11s %-11s %-11s %-11s\n", "Sample", "Frames",
	   "Loop start", "Loop end", "Unknown");
    frame_cnt = 0;
    for(i = 0; i < bin->samples; i++)
    {
	mrs8_sample_def_t *sd = &bin->sample_def[i];

	printf("%-7d %-11d %-11d %-11d %-11d\n", i, sd->frames,
		sd->frame_loop_begin, sd->frame_loop_end, sd->unknown1);
	frame_cnt += sd->frames;
    }
    printf("Total frames:%d\n", frame_cnt);
}

static void MRS8_export_samples(mrs8_bin_t *bin)
{
    int		 i;
    short	*sample;

    sample = bin->sample_data;
    for(i = 0; i < bin->samples; i++)
    {
	mrs8_sample_def_t *sd = &bin->sample_def[i];
	char file[80];
	FILE *F;

	sprintf(file, "sample%2.2d.wav", i);
	printf("Exporting %s...", file);
	F = fopen(file, "w+b");
	if (F == NULL)
	{
	    fprintf(stderr, "Unable to open output file %s\n", file);
	    return;
	}

	write_wave_header(F, sd->frames, 1, 44100, 16);
	fwrite(sample, sd->frames, sizeof(*sample), F);
	sample += sd->frames;
	fclose(F);
	printf("done\n");
    }
}

static void MRS8_import_samples(mrs8_bin_t *bin)
{
    int			 i;
    short		*sample;
    mrs8_sample_def_t	*sd = NULL;

    bin->modified = FALSE;
    sample = bin->sample_data;
    bin->total_frames = 0;
    for(i = 0; i < bin->samples; i++)
    {
	char file[80];
	FILE *F;
	WaveHeader header_chunk;
	DataChunk data_chunk;

	sd = &bin->sample_def[i];
	sprintf(file, "sample%2.2d.wav", i);
	printf("Importing %s...", file);
	F = fopen(file, "rb");
	if (F == NULL)
	{
	    fprintf(stderr, "Unable to open input file %s\n", file);
	    return;
	}

	read_wave_header(F, &header_chunk, &data_chunk);
	if (header_chunk.fmt.format != WAVE_FORMAT_PCM ||
	    header_chunk.fmt.channels != 1 ||
	    header_chunk.fmt.sample_rate != 44100 ||
	    header_chunk.fmt.bits_per_sample != 16)
	{
	    printf("Bad format %s: wav file must be 16 bit mono 44100Hz\n",
		   file);
	    return;
	}

	/* FIXME: let user specify loop points for bass samples... */
	sd->frames = data_chunk.chunk_size / sizeof(*sample);
	if (sd->frames == 0)
	{
	    sd->frame_loop_begin = 0;
	    sd->frame_loop_end = 0;
	    sd->unknown1 = 0;
	}
	if (i < 47)
	    sd->frame_loop_end = sd->frames;
#if 0
	if (i < 47)
	{
	    sd->frame_loop_begin = 0;
	    sd->unknown1 = 0;
	}
#endif

	/* FIXME: verify we havn't overrun the sample space... */

	fread(sample, sd->frames, sizeof(*sample), F);
	sample += sd->frames;
	bin->total_frames += sd->frames;
	fclose(F);
	printf("done\n");
    }

    printf("Frames:%d of 1483689\n", bin->total_frames);
    if (bin->total_frames > 1483689)
    {
	printf("Too many frames, this file will not flash successfully.\n"
	       "Reduce the amount of sample data and try again.");
    }

    if (bin->total_frames < 1483689)
    {
	int pad_frames = 1483689-bin->total_frames;

	printf("Padding last wav with %d frames\n", pad_frames);
	memset(sample, 0, pad_frames * sizeof(*sample));
	sd->frames += pad_frames;
	bin->total_frames += pad_frames;
    }
    printf("Total frames:%d\n", bin->total_frames);
    bin->modified = TRUE;
}

int32_t getsum(int8_t *buf, int print, uint32_t start, uint32_t end)
{
    int		 i;

    MRS8_chksum(0x00, 1);
    for(i = start; i <= end; i++)
	MRS8_chksum(buf[i], 0);
    if (print)
	printf("Region: 0x%8.8x - 0x%8.8x Sum:0x%8.8x\n", start, end,
	       MRS8_chksum(0, 0));

    return MRS8_chksum(0, 0);
}

int main(int argc, char *argv[])
{
    int		 opt;
    int		 cmd;

    static struct option long_options[] =
    {
	{"help",		0, 0, '?'},

	{"dump-info",		0, 0, 'd'},
	{"export-samples",	0, 0, 'x'},
	{"import-samples",	0, 0, 'i'},

	{0,0,0,0},
    };


    if (argc < 2)
	usage(argv[0]);

    cmd = 0;
    while ((opt = getopt_long(argc, argv, "dxi", long_options, NULL)) != -1)
    {
	switch(opt)
	{
	default:
	case '?':
	    usage(argv[0]);
	    break;
	case 'd':
	    cmd |= CMD_DUMP_INFO;
	    break;
	case 'x':
	    cmd |= CMD_EXPORT_SAMPLES;
	    break;
	case 'i':
	    cmd |= CMD_IMPORT_SAMPLES;
	    break;
	}
    }

    if (optind == argc)
	usage(argv[0]);

    bin = MRS8_bin_open(argv[optind]);
    if (bin == NULL)
    {
	fprintf(stderr, "Unable to open %s\n", argv[optind]);
	exit(EXIT_FAILURE);
    }

    if (cmd & CMD_DUMP_INFO)
	MRS8_dump_info(bin);

    if (cmd & CMD_EXPORT_SAMPLES)
	MRS8_export_samples(bin);

    if (cmd & CMD_IMPORT_SAMPLES)
	MRS8_import_samples(bin);

    if (bin->modified)
	MRS8_bin_write(bin);
    MRS8_bin_close(bin);

    int8_t *b;
    FILE *F;
    b = malloc(0x400000);
    F = fopen(argv[optind], "r+b");
    fread(b, 0x400000, 1, F);
    fclose(F);

    getsum(b, 1, 0x0, 0x3ffffb);
    getsum(b, 1, 0x2, 0x3ffffb);
    getsum(b, 1, 0x4, 0x3ffffb);
    getsum(b, 1, 0x8, 0x3ffffb);
    getsum(b, 1, 0x020000, 0x3ffffb);
    getsum(b, 1, 0x020200, 0x3ffffb);
    getsum(b, 1, 0x020300, 0x3ffffb);

    getsum(b, 1, 0x020508,  0x2050b);
    getsum(b, 1, 0x020508,  0x2050b);
    getsum(b, 1, 0x020518,  0x3ebff);
    getsum(b, 1, 0x03ec0c,  0x847ab);
    getsum(b, 1, 0x0847b8,  0x84ab3);
    getsum(b, 1, 0x127cc6, 0x12b8a9);

    getsum(b, 1, 0x84ab8, 0x3ffffb);
    getsum(b, 1, 0x84ac0, 0x3ffffb);
    getsum(b, 1, 0xe5412, 0x3ffffb);
    getsum(b, 1, 0xee970, 0x3ffffb);
    getsum(b, 1, 0x127cc6, 0x3ffffb);
    getsum(b, 1, 0x127ec6, 0x3ffffb);

    int i;
#if 0
    for(i = 0; i < 0x3ffffb; i++)
    {
	if (getsum(b, 0, i, 0x3ffffb) == 0x24066)
	    printf("\nMatch! Offset: 0x%8.8x\n", i);
	if ((i % 17) == 0)
	{
	    printf("Offset 0x%8.8x\r", i);
	    fflush(stdout);
	}
    }
    printf("\nDone\n");
#endif
    int32_t sum = 0;
    for(i = 0; i < 0x3ffffc; i++)
    {
	sum += b[i];
    }
    printf("Sum: 0x%8.8x\n", sum);

    exit(EXIT_SUCCESS);
}
