/*
 * PS02_to_PS04
 *
 * Copyright (C) 2003-2004 Dwight Engen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <libzoom.h>


#define SONG_LIST		0x000000001
#define SONG_DUMP		0x000000002
#define SONG_DOUBLE		0x000000004
#define SONG_COPY_MEASURES	0x000000008
#define PTN_LIST		0x000000010
#define PTN_DUMP		0x000000020

static ps02_pts_t *pts;

static void PS02_pattern_dump(ps02_pattern_t *ptn)
{
    int i;
    int ppq = 0;

    printf("Pattern  : %6.6s\n", ptn->name);
    printf("Beats    : %d\n", ptn->beats);
    printf("Bars     : %d\n", ptn->bars);
    printf("Events   : %d\n", ptn->events);

    for (i = 0; i < ptn->events; i++)
    {
	ptn_event_t *ev = &ptn->event[i];

        ppq += ev->ppq_delta;
	printf("Ev:%3d PPQDelta:%2d PPQDuration:%3d ", i+1, ev->ppq_delta,
		ev->ppq_duration);
	switch(ev->type)
	{
	case PTN_EVENT_DRUM:		printf("EvTyp:Drum      "); break;
	case PTN_EVENT_DRUM_BANK:	printf("EvTyp:Drum bank "); break;
	case PTN_EVENT_BASS:		printf("EvTyp:Bass      "); break;
	case PTN_EVENT_BASS_BANK:	printf("EvTyp:Bass bank "); break;
	case PTN_EVENT_PPQ:		printf("EvTyp:PTN PPQ   "); break;
	case PTN_EVENT_END:		printf("EvTyp:PTN END   "); break;
	}

	printf("Vel:%2.2d MIDI:%2.2d ", ev->velocity15, ev->midi_note);

	switch(ev->type)
	{
	case PTN_EVENT_DRUM:
	    printf("%s\n", PS0x_midi_drum_name(ev->midi_note));
	    break;
	case PTN_EVENT_BASS:
	    printf("%s %s\n", PS0x_midi_note_name(ev->midi_note), ev->velocity15 ? "" : "off");
	    break;
	default:
	    printf("\n");
	    break;
	}
    }
    printf("       PPQTotal:%d\n", ppq);
}



static void PS02_song_dump(ps02_song_t *song)
{
    int i;
    int j;

    printf("Song    : %6.6s\n", song->name);
    printf("Measures: %d\n", song->measures);
    printf("Tempo   : %d\n", song->tempo);
    printf("Drum Vol: %d\n", song->drum_volume);
    printf("Bass Vol: %d\n", song->bass_volume);

    for (i = 0; i < song->measures; i++)
    {
	printf("Meas:%3d Ptn:%6.6s (0x%2.2x) ", i+1,
		pts->rhy_pattern[song->measure[i].pattern]->name,
		song->measure[i].pattern);

	for (j = 0; j < song->measure[i].beats; j++)
	{
	    unsigned char note = song->measure[i].note[j];

	    printf("%d:", j);
	    if (note == 0xff)
	    {
		printf("      ");
		continue;
	    }

	    printf("%-2s%3s ", note_root_name[note & 0x0f],
			    note_tone_name[(note & 0xf0) >> 4]);
	}
	printf("\n");
    }
    printf("\n");
}



static void usage(char *prgname)
{
    printf("ps02_tool v1.00 Copyright (C) 2003-2004 Dwight Engen\n\n");
    printf("This program is free software; you can redistribute it and/or modify\n"
	   "it under the terms of the GNU General Public License as published by\n"
	   "the Free Software Foundation; either version 2 of the License, or\n"
	   "(at your option) any later version.\n\n");

    printf("Usage: %s <OPTION> <file.pts>\n\n", prgname);
    printf(" -h  --help\n\n");

    printf(" -s  --list-songs\n");
    printf(" -d  --dump-song=<song #>\n");
    printf(" -u  --double-song=<song #>\n");
    printf(" -c  --copy-measures=<song #,begin measure,end measure,dest measure>\n\n");

    printf(" -p  --list-patterns\n");
    printf(" -t  --dump-pattern=<ptn #>\n");
    exit(EXIT_FAILURE);
}



int main(int argc, char *argv[])
{
    int		 songnr = 1;
    int		 ptnnr = 1;
    int		 opt;
    int		 i;
    int		 cmd;
    int		 start = 1, end = 1, dest = 1;

    static struct option long_options[] =
    {
	{"help",		0, 0, '?'},

	{"list-songs",		0, 0, 's'},
	{"dump-song",		1, 0, 'd'},
	{"double-song",		1, 0, 'u'},
	{"copy-measures",	1, 0, 'c'},

	{"list-patterns",	0, 0, 'p'},
	{"dump-pattern",	0, 0, 't'},
	{0,0,0,0},
    };


    if (argc < 2)
	usage(argv[0]);

    cmd = 0;
    while ((opt = getopt_long(argc, argv, "sd:u:c:pt:v:", long_options, NULL)) != -1)
    {
	switch(opt)
	{
	default:
	case '?':
	    usage(argv[0]);
	    break;
	case 's':
	    cmd |= SONG_LIST;
	    break;
	case 'd':
	    if (!optarg)
		usage(argv[0]);
	    songnr = strtoul(optarg, NULL, 0);
	    if (songnr > 100)
	    {
		printf("Invalid song %d (1 - 100 are valid)\n", songnr);
		break;
	    }
	    cmd |= SONG_DUMP;
	    break;
	case 'u':
	    if (!optarg)
		usage(argv[0]);
	    songnr = strtoul(optarg, NULL, 0);
	    cmd |= SONG_DOUBLE;
	    break;
	case 'c':
	  {
	    char *tok;

	    tok = strtok(optarg, ",");
	    if (!tok)
		usage(argv[0]);
	    songnr = strtoul(tok, NULL, 0);

	    tok = strtok(NULL, ",");
	    if (!tok)
		usage(argv[0]);
	    start = strtoul(tok, NULL, 0);

	    tok = strtok(NULL, ",");
	    if (!tok)
		usage(argv[0]);
	    end = strtoul(tok, NULL, 0);

	    tok = strtok(NULL, ",");
	    if (!tok)
		usage(argv[0]);
	    dest = strtoul(tok, NULL, 0);
	    cmd |= SONG_COPY_MEASURES;
	    break;
	  }

	case 'p':
	    cmd |= PTN_LIST;
	    break;
	case 't':
	    if (!optarg)
		usage(argv[0]);
	    ptnnr = strtoul(optarg, NULL, 0);
	    cmd |= PTN_DUMP;
	    break;
	}
    }

    if (optind == argc)
	usage(argv[0]);

    pts = PS02_pts_open(argv[optind]);
    if (pts == NULL)
    {
	fprintf(stderr, "Unable to open %s\n", argv[optind]);
	exit(EXIT_FAILURE);
    }

    if (cmd & SONG_LIST)
    {
	for (i=0; i < 100; i++)
	    printf("Song:%3d %6.6s\n", i+1, pts->song[i]->name);
    }

    if (cmd & SONG_DUMP)
	PS02_song_dump(pts->song[songnr-1]);

    if (cmd & SONG_DOUBLE)
	PS02_song_double(pts->song[songnr-1]);

    if (cmd & SONG_COPY_MEASURES)
	PS02_song_copy_measures(pts->song[songnr-1], start-1, end-1, dest-1);

    if (cmd & PTN_LIST)
    {
	for (i = 0; i < pts->rhy_pattern_cnt; i++)
	    printf("Ptn:%3d %6.6s\n", i+1, pts->rhy_pattern[i]->name);
    }

    if (cmd & PTN_DUMP)
    {
	if (ptnnr > pts->rhy_pattern_cnt)
	    printf("Invalid pattern %d (%d - %ld are valid)\n", ptnnr,
		   1, pts->rhy_pattern_cnt);
	else
	    PS02_pattern_dump(pts->rhy_pattern[ptnnr-1]);
    }

    PS02_pts_write(pts);
    PS02_pts_close(pts);
    exit(EXIT_SUCCESS);
}


#if 0
    seq = PS04_seq_open("Rhythm01.seq");
    if (seq == NULL)
    {
	fprintf(stderr, "Unable to open Rhythm01.seq\n");
	exit(1);
    }
    PS04_seq_write(seq);
    PS04_seq_close(seq);


    INPCH = fopen("User.pch", "r+b");
    if (INPCH == NULL)
    {
	fprintf(stderr, "Unable to open User.pch\n");
	exit(1);
    }

    pch = malloc(sizeof(*pch));
    memset(pch, 0, sizeof(*pch));

    PS02_pch_read(INPCH, pch);
#endif
