/*
 * PS02_to_PS04
 *
 * Copyright (C) 2003-2004 Dwight Engen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <libzoom.h>


#define SONG_LIST		0x000000001
#define SONG_DUMP		0x000000002
#define SONG_DOUBLE		0x000000004
#define SONG_COPY_MEASURES	0x000000008
#define PTN_LIST		0x000000010
#define PTN_DUMP		0x000000020
#define EVENT_DELETE		0x000000040

static ps04_seq_t *seq;

static const char *PS04_event_name(song_event_type_t event)
{
    static const char event_name[][9] =
    {
	"Invalid",
	"Pattern",
	"Note",
	"Tempo",
	"Timesig",
	"Drum Vol",
	"Bass Vol",
	"Drum Kit",
	"Bass Kit",
    };
    if (event < 9)
	return event_name[event];
    else
	return "Unknown";
}



static void PS04_song_dump(ps04_song_t *song, ps04_seq_t *seq)
{
    int i;
    int ppq;

    printf("Song    : %-8.8s\n", song->name);
    printf("Events  : %d\n", song->events);

    ppq = 0;
    for (i = 0; i < song->events; i++)
    {
	unsigned short data = song->event[i].data;

	printf("Event:%4.4d PPQ:%4.4d (%5.5d) Meas:%3.3d Type:%-10s ",
		i+1, song->event[i].next_ppq, ppq, ppq/(PS04_PPQ*4)+1,
		PS04_event_name(song->event[i].type));

	switch(song->event[i].type)
	{
	case SONG_EVENT_PATTERN:
	    printf("%-8.8s (0x%3.3x)\n", seq->pattern[data]->name, data);
	    break;
	case SONG_EVENT_NOTE:
	    printf("%-2s%3s\n", note_root_name[data & 0x0f],
		   note_tone_name[(data & 0xf0) >> 4]);
	    break;

	case SONG_EVENT_TEMPO:
	    printf("%d.%d\n", data/10, data - ((data/10)*10));
	    break;

	case SONG_EVENT_TIMESIG:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_DRUM_VOL:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_BASS_VOL:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_DRUM_KIT:
	    printf("%d\n", data);
	    break;
	case SONG_EVENT_BASS_KIT:
	    printf("%d\n", data);
	    break;
	default:
	    printf("Unknown event type!\n");
	}

	/* Handle large/small PPQ's */
	if (song->event[i].next_ppq == 0xff)
	    ppq += *(uint32_t *)&song->event[++i];
	else
	    ppq += song->event[i].next_ppq;
    }
}



static void PS04_pattern_dump(ps04_pattern_t *ptn)
{
    int i;
    int ppq = 0;

    printf("Pattern  : %6.6s\n", ptn->name);
    printf("Beats    : %d\n", ptn->beats);
    printf("Bars     : %d\n", ptn->bars);
    printf("Drum Kit : %d\n", ptn->drum_kit);
    printf("Bass Kit : %d\n", ptn->bass_kit);
    printf("Drum Lvl : %d\n", ptn->drum_level);
    printf("Bass Lvl : %d\n", ptn->bass_level);
    printf("Events   : %d\n", ptn->sevents);

    for (i = 0; i < ptn->sevents; i++)
    {
	ptn_event_t *ev = &ptn->event[i];

        ppq += ev->ppq_delta;
	printf("Ev:%3d PPQDelta:%2d PPQDuration:%3d ", i+1, ev->ppq_delta,
		ev->ppq_duration);
	switch(ev->type)
	{
	case PTN_EVENT_DRUM:		printf("EvTyp:Drum      "); break;
	case PTN_EVENT_DRUM_BANK:	printf("EvTyp:Drum bank "); break;
	case PTN_EVENT_BASS:		printf("EvTyp:Bass      "); break;
	case PTN_EVENT_BASS_BANK:	printf("EvTyp:Bass bank "); break;
	case PTN_EVENT_PPQ:		printf("EvTyp:PTN PPQ   "); break;
	case PTN_EVENT_END:		printf("EvTyp:PTN END   "); break;
	}

	printf("Vel:%2.2d MIDI:%2.2d ", ev->velocity15, ev->midi_note);

	switch(ev->type)
	{
	case PTN_EVENT_DRUM:
	    printf("%s\n", PS0x_midi_drum_name(ev->midi_note));
	    break;
	case PTN_EVENT_BASS:
	    printf("%s %s\n", PS0x_midi_note_name(ev->midi_note), ev->velocity15 ? "" : "off");
	    break;
	default:
	    printf("\n");
	    break;
	}
    }
    printf("       PPQTotal:%d\n", ppq);
}



static void usage(char *prgname)
{
    printf("ps04_tool v1.00 Copyright (C) 2003-2005 Dwight Engen\n\n");
    printf("This program is free software; you can redistribute it and/or modify\n"
	   "it under the terms of the GNU General Public License as published by\n"
	   "the Free Software Foundation; either version 2 of the License, or\n"
	   "(at your option) any later version.\n\n");

    printf("Usage: %s <OPTION> <file.seq>\n\n", prgname);
    printf(" -h  --help\n\n");

    printf(" -s  --list-songs\n");
    printf(" -d  --dump-song=<song #>\n");
    printf(" -u  --double-song=<song #>\n");
    printf(" -c  --copy-measures=<song #,begin measure,end measure,dest measure>\n");
    printf(" -e  --event-delete=<event type #>\n\n");

    printf(" -p  --list-patterns\n");
    printf(" -t  --dump-pattern=<ptn #>\n");
    exit(EXIT_FAILURE);
}



int main(int argc, char *argv[])
{
    int		 songnr = 1;
    int		 ptnnr = 1;
    int		 eventnr = 1;
    int		 opt;
    int		 i;
    int		 cmd;
    int		 start = 1, end = 1, dest = 1;

    static struct option long_options[] =
    {
	{"help",		0, 0, '?'},

	{"list-songs",		0, 0, 's'},
	{"dump-song",		1, 0, 'd'},
	{"double-song",		1, 0, 'u'},
	{"copy-measures",	1, 0, 'c'},
	{"event-delete",	1, 0, 'e'},

	{"list-patterns",	0, 0, 'p'},
	{"dump-pattern",	0, 0, 't'},
	{0,0,0,0},
    };


    if (argc < 2)
	usage(argv[0]);

    cmd = 0;
    while ((opt = getopt_long(argc, argv, "sd:u:c:e:pt:v:", long_options, NULL)) != -1)
    {
	switch(opt)
	{
	default:
	case '?':
	    usage(argv[0]);
	    break;
	case 's':
	    cmd |= SONG_LIST;
	    break;
	case 'd':
	    if (!optarg)
		usage(argv[0]);
	    songnr = strtoul(optarg, NULL, 0);
	    if (songnr > 10)
	    {
		printf("Invalid song %d (1 - 10 are valid)\n", songnr);
		break;
	    }
	    cmd |= SONG_DUMP;
	    break;
	case 'e':
	    if (!optarg)
		usage(argv[0]);
	    eventnr = strtoul(optarg, NULL, 0);
	    cmd |= EVENT_DELETE;
	    break;
	case 'u':
	    if (!optarg)
		usage(argv[0]);
	    songnr = strtoul(optarg, NULL, 0);
	    cmd |= SONG_DOUBLE;
	    break;
	case 'c':
	  {
	    char *tok;

	    tok = strtok(optarg, ",");
	    if (!tok)
		usage(argv[0]);
	    songnr = strtoul(tok, NULL, 0);

	    tok = strtok(NULL, ",");
	    if (!tok)
		usage(argv[0]);
	    start = strtoul(tok, NULL, 0);

	    tok = strtok(NULL, ",");
	    if (!tok)
		usage(argv[0]);
	    end = strtoul(tok, NULL, 0);

	    tok = strtok(NULL, ",");
	    if (!tok)
		usage(argv[0]);
	    dest = strtoul(tok, NULL, 0);
	    cmd |= SONG_COPY_MEASURES;
	    break;
	  }

	case 'p':
	    cmd |= PTN_LIST;
	    break;
	case 't':
	    if (!optarg)
		usage(argv[0]);
	    ptnnr = strtoul(optarg, NULL, 0);
	    cmd |= PTN_DUMP;
	    break;
	}
    }

    if (optind == argc)
	usage(argv[0]);

    seq = PS04_seq_open(argv[optind]);
    if (seq == NULL)
    {
	fprintf(stderr, "Unable to open %s\n", argv[optind]);
	exit(EXIT_FAILURE);
    }

    if (cmd & SONG_LIST)
    {
	for (i=0; i < 10; i++)
	    printf("Song:%3d %-6.6s\n", i+1, seq->song[i]->name);
    }

    if (cmd & SONG_DUMP)
	PS04_song_dump(seq->song[songnr-1], seq);

#if 0
    if (cmd & SONG_DOUBLE)
	PS02_song_double(pts->song[songnr-1]);

    if (cmd & SONG_COPY_MEASURES)
	PS02_song_copy_measures(pts->song[songnr-1], start-1, end-1, dest-1);
#endif

    if (cmd & PTN_LIST)
    {
	for (i = 0; i < seq->patterns; i++)
	    printf("Ptn:%3d %-6.6s\n", i+1, seq->pattern[i]->name);
    }

    if (cmd & EVENT_DELETE)
    {
	printf("Deleting all events %s from song:%d\n", PS04_event_name(eventnr), songnr);
	ps04_song_t *song = seq->song[songnr-1];
	for (i = 0; i < song->events; )
	{
	    if (song->event[i].type == eventnr)
	    {
		song->event[i-1].next_ppq += song->event[i].next_ppq;
		memmove(&song->event[i], &song->event[i+1],
			(song->events - i - 1) * sizeof(song->event[0]));
		song->events--;
	    }
	    else
		i++;
	}
    }

    if (cmd & PTN_DUMP)
    {
	if (ptnnr > seq->patterns)
	    printf("Invalid pattern %d (%d - %ld are valid)\n", ptnnr,
		   1, seq->patterns);
	else
	    PS04_pattern_dump(seq->pattern[ptnnr-1]);
    }

    PS04_seq_write(seq);
    PS04_seq_close(seq);
    exit(EXIT_SUCCESS);
}
