/*
 * MRS8 - Manage the data of a Zoom MRS-8.
 *
 * Copyright (C) 2003-2005 Dwight Engen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: $
 */

#include <dirent.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include "mrs8.h"
#include <libzoom.h>


typedef struct
{
    int		 number;
    int		 vtake;
    char	 name[9];
    char	 filename[NAME_MAX];
} track_t;

typedef struct
{
    int		 number;
    int		 on_card;
    char	 filename[NAME_MAX];
    char	 name[9];
    track_t	*track[MAX_TRACKS][MAX_VTAKES];
} project_t;

static project_t	*project[MAX_PROJECTS];
static char		 mrs8path[PATH_MAX+1];
static GtkWidget	*WinMain = NULL;
static GtkWidget	*WinProgress = NULL;
static gboolean		 ignore_changed;

enum 
{
    PROJECT_COL_DATA,
    PROJECT_COL_NUMBER,
    PROJECT_COL_NUMBER_TEXT,
    PROJECT_COL_NAME,
    PROJECT_COLS
};

enum 
{
    TRACK_COL_DATA,
    TRACK_COL_NUMBER_TEXT,
    TRACK_COL_VTAKE,
    TRACK_COL_NAME,
    TRACK_COLS
};

static void
gui_window_title (GtkWidget *win, const char *fmt,...)
{
    gchar title[256];
    va_list v_lst;

    va_start(v_lst, fmt);
    vsnprintf(title, sizeof(title)-1, fmt, v_lst);
    va_end(v_lst);
    gtk_window_set_title(GTK_WINDOW(win), title);
}

static void *
gui_get (GtkWidget *widget, const gchar *widget_name)
{
    GtkWidget  *parent, *found_widget;

    for (;;)
    {
	if (GTK_IS_MENU(widget))
	    parent = gtk_menu_get_attach_widget(GTK_MENU(widget));
	else
	    parent = widget->parent;
	if (!parent)
	    parent = g_object_get_data(G_OBJECT(widget), "GladeParentKey");
	if (parent == NULL)
	    break;
	widget = parent;
    }

    found_widget = (GtkWidget *) g_object_get_data(G_OBJECT(widget),
						   widget_name);
    if (!found_widget)
	g_warning("Widget not found: %s", widget_name);
    return found_widget;
}

static void
gui_on_modal_ok (GtkButton *button, gboolean *user_data)
{
    *user_data = TRUE;
    gtk_main_quit ();
}

static void
gui_on_modal_cancel (GtkButton *button, gboolean *user_data)
{
    *user_data = FALSE;
    gtk_main_quit ();
}

static gboolean
gui_on_modal_destroy (GtkWidget *widget, GdkEvent *event, gboolean *user_data)
{
    *user_data = FALSE;
    gtk_main_quit ();
    return FALSE;
}

gboolean
gui_modal_ask (gboolean can_cancel, const char *title,
	       const char *ok, const char *cancel,
	       const char *fmt,...)
{
    GtkWidget	*MDLGVBox;
    GtkWidget	*MDLGLabel;
    GtkWidget	*separator;
    GtkWidget	*HButtonBox;
    GtkWidget	*Ok;
    GtkWidget	*Cancel;
    GtkWidget	*win_modal_ask;
    char	msg_buf[512];
    va_list	v_lst;
    gboolean	rc;

    win_modal_ask = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (win_modal_ask), title);
    gtk_window_set_modal (GTK_WINDOW (win_modal_ask), TRUE);
    gtk_window_set_policy (GTK_WINDOW (win_modal_ask), FALSE, TRUE, TRUE);

    MDLGVBox = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (MDLGVBox);
    gtk_container_add (GTK_CONTAINER (win_modal_ask), MDLGVBox);
    gtk_container_set_border_width (GTK_CONTAINER (MDLGVBox), 2);

    va_start(v_lst, fmt);
    vsnprintf(msg_buf, sizeof(msg_buf)-1, fmt, v_lst);
    va_end(v_lst);

    MDLGLabel = gtk_label_new (msg_buf);
    gtk_widget_show (MDLGLabel);
    gtk_box_pack_start (GTK_BOX (MDLGVBox), MDLGLabel, TRUE, TRUE, 0);
    gtk_label_set_justify (GTK_LABEL (MDLGLabel), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment (GTK_MISC (MDLGLabel), 0.5, 0.10);

    separator = gtk_hseparator_new ();
    gtk_widget_show (separator);
    gtk_box_pack_start (GTK_BOX (MDLGVBox), separator, FALSE, FALSE, 4);

    HButtonBox = gtk_hbutton_box_new ();
    gtk_widget_show (HButtonBox);
    gtk_box_pack_end (GTK_BOX (MDLGVBox), HButtonBox, FALSE, FALSE, 0);
    gtk_button_box_set_layout (GTK_BUTTON_BOX (HButtonBox), GTK_BUTTONBOX_END);
    gtk_button_box_set_spacing (GTK_BUTTON_BOX (HButtonBox), 5);

    Ok = gtk_button_new_with_label (ok);
    gtk_widget_show (Ok);
    gtk_container_add (GTK_CONTAINER (HButtonBox), Ok);
    GTK_WIDGET_SET_FLAGS (Ok, GTK_CAN_DEFAULT);

    if (can_cancel)
    {
	Cancel = gtk_button_new_with_label (cancel);
	gtk_widget_show (Cancel);
	gtk_container_add (GTK_CONTAINER (HButtonBox), Cancel);
	GTK_WIDGET_SET_FLAGS (Cancel, GTK_CAN_DEFAULT);

	gtk_signal_connect (GTK_OBJECT (Cancel), "clicked",
			    GTK_SIGNAL_FUNC (gui_on_modal_cancel),
			    &rc);
    }

    gtk_signal_connect (GTK_OBJECT (win_modal_ask), "destroy_event",
			GTK_SIGNAL_FUNC (gui_on_modal_destroy),
			&rc);
    gtk_signal_connect (GTK_OBJECT (win_modal_ask), "delete_event",
			GTK_SIGNAL_FUNC (gui_on_modal_destroy),
			&rc);
    gtk_signal_connect (GTK_OBJECT (Ok), "clicked",
			GTK_SIGNAL_FUNC (gui_on_modal_ok),
			&rc);

    gtk_widget_grab_default (Ok);
    gtk_widget_show(win_modal_ask);
    gtk_main();

    gtk_widget_destroy(win_modal_ask);
    return rc;
}

static gboolean
gui_modal_filename_ask(char *title, char *file)
{
    GtkWidget	*fs_win;
    gboolean	 rc = FALSE;

    fs_win = gtk_file_selection_new(title);
    gtk_window_set_modal(GTK_WINDOW(fs_win), TRUE);
    gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(fs_win));
    gtk_signal_connect(GTK_OBJECT(fs_win), "destroy_event",
			GTK_SIGNAL_FUNC(gui_on_modal_destroy), &rc);
    gtk_signal_connect(GTK_OBJECT(fs_win), "delete_event",
			GTK_SIGNAL_FUNC(gui_on_modal_destroy), &rc);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fs_win)->ok_button),
			"clicked", GTK_SIGNAL_FUNC(gui_on_modal_ok), &rc);
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fs_win)->cancel_button),
			"clicked", GTK_SIGNAL_FUNC(gui_on_modal_cancel), &rc);

    gtk_widget_show (fs_win);
    gtk_main ();

    if (rc)
	strcpy (file, gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs_win)));

    gtk_widget_destroy (GTK_WIDGET(fs_win));
    return rc;
}

void
gui_sync(void)
{
    while(g_main_iteration(FALSE))
	;
}




gint
sort_num_func (GtkTreeModel *model, GtkTreeIter *i1, GtkTreeIter *i2, gpointer userdata)
{
    guint a, b, column;

    column = GPOINTER_TO_INT(userdata);
    gtk_tree_model_get(model, i1, column, &a, -1);
    gtk_tree_model_get(model, i2, column, &b, -1);

    if (a != b)
	return (a > b) ? 1 : -1;
    return 0;
}

/* Track handling functions */

gint
sort_track_func (GtkTreeModel *model, GtkTreeIter *i1, GtkTreeIter *i2, gpointer userdata)
{
    track_t *trk1;
    track_t *trk2;

    gtk_tree_model_get(model, i1, TRACK_COL_DATA, &trk1, -1);
    gtk_tree_model_get(model, i2, TRACK_COL_DATA, &trk2, -1);

    if (trk1->number == trk2->number)
	return trk1->vtake > trk2->vtake ? 1 : -1;

    return trk1->number > trk2->number ? 1 : -1;
}

static int track_char_to_num(char ch)
{
    switch(ch)
    {
    case 'l':
    case 'L':
	return 9;
	break;
    case 'r':
    case 'R':
	return 10;
	break;
    default:
	return ch - '0';
    }
}

static char track_num_to_char(int number)
{
    switch(number)
    {
    case 9:
	return 'l';
    case 10:
	return 'r';
    default:
	return '0' + number;
    }
}

track_t *track_new(int number, int vtake)
{
    track_t *trk;

    trk = malloc(sizeof(*trk));
    memset(trk, 0, sizeof(*trk));
    trk->number = number;
    trk->vtake  = vtake;
    sprintf(trk->filename, "track%c_%d.zaf",
	    track_num_to_char(number), vtake);
    return trk;
}

static void track_get_name(project_t *prj, track_t *trk)
{
    gchar	*zaf;
    FILE	*F;

    zaf = g_build_filename(mrs8path, prj->filename, trk->filename, NULL);
    F = fopen(zaf, "rb");
    if (F == NULL)
    {
	trk->name[0] = '\0';
	goto out;
    }

    fseek(F, 1024, SEEK_SET);
    fread(trk->name, 8, 1, F);
    trk->name[sizeof(trk->name)-1] = '\0';
    fclose(F);
out:
    g_free(zaf);
}

static int tracks_selected(track_t *track[])
{
    GtkTreeSelection	*sel;
    GtkTreeModel	*model;
    GtkTreeIter		 row;
    GList		*list;
    GList		*iter;
    int			 i;

    sel = gtk_tree_view_get_selection(gui_get(WinMain, "TrackTreeView"));
    list = gtk_tree_selection_get_selected_rows(sel, &model);
    
    for(iter = list, i = 0; iter; iter = g_list_next(iter), i++)
    {
	gtk_tree_model_get_iter(model, &row, iter->data);
	gtk_tree_model_get(model, &row, TRACK_COL_DATA, &track[i], -1);
    }

    g_list_foreach(list, (GFunc)gtk_tree_path_free, NULL);
    g_list_free(list);
    return i;
}

static void track_update(project_t *prj, track_t *trk, GtkTreeIter *row)
{
    GtkListStore	*ls;
    char		 buf[40];

    ls = GTK_LIST_STORE(gtk_tree_view_get_model(gui_get(WinMain, "TrackTreeView")));
    //path = gtk_tree_path_new_from_indices(track->number * 10 + track->vtake, -1);
    //gtk_tree_model_get_iter(GTK_TREE_MODEL(ls), &row, path);

    if (trk->number < 9 || trk->number > 10)
	sprintf(buf, "%2.2d", trk->number);
    if (trk->number == 9)
	strcpy(buf, "Master Left");
    if (trk->number == 10)
	strcpy(buf, "Master Right");

    gtk_list_store_set(ls, row, TRACK_COL_DATA, trk, -1);

    gtk_list_store_set(ls, row, TRACK_COL_NUMBER_TEXT, buf, -1);

    sprintf(buf, "%d", trk->vtake);
    gtk_list_store_set(ls, row, TRACK_COL_VTAKE, buf, -1);
    
    track_get_name(prj, trk);
    gtk_list_store_set(ls, row, TRACK_COL_NAME, trk->name, -1);
    //gtk_tree_path_free(path);
}

static void track_list_populate(project_t *prj)
{
    GtkListStore	*ls;
    GtkTreeView		*tv;
    GtkTreeIter		 row;
    GtkTreeViewColumn	*col;
    GtkTreeSelection	*sel;
    GtkCellRenderer	*rend;

    tv = gui_get(WinMain, "TrackTreeView");
    ls = GTK_LIST_STORE(gtk_tree_view_get_model(tv));
    if (ls == NULL)
    {
	ls = gtk_list_store_new(TRACK_COLS, G_TYPE_POINTER, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	gtk_tree_view_set_model(tv, GTK_TREE_MODEL(ls));

	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(ls), TRACK_COL_NUMBER_TEXT,
					sort_track_func, GINT_TO_POINTER(TRACK_COL_NUMBER_TEXT), NULL);

	rend = gtk_cell_renderer_text_new();
	col  = gtk_tree_view_column_new_with_attributes("Track", rend, "text",
							TRACK_COL_NUMBER_TEXT, NULL);
	gtk_tree_view_append_column(tv, col);

	col  = gtk_tree_view_column_new_with_attributes("VTake", rend, "text",
							TRACK_COL_VTAKE, NULL);
	gtk_tree_view_append_column(tv, col);

	col  = gtk_tree_view_column_new_with_attributes("Name", rend, "text",
							TRACK_COL_NAME, NULL);
	gtk_tree_view_append_column(tv, col);

	sel = gtk_tree_view_get_selection(tv);
	gtk_tree_selection_set_mode(sel, GTK_SELECTION_MULTIPLE);
//	g_signal_connect(G_OBJECT(sel), "changed",
//			 G_CALLBACK(project_selected), NULL);
    }

#if 0
    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(ls),
					 GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID,
					 GTK_SORT_ASCENDING);
#endif

    gtk_list_store_clear(ls);

    {
	DIR		*dir;
	gchar		*dirname;

	dirname = g_build_filename(mrs8path, prj->filename, NULL);
	dir = opendir(dirname);
	if (dir)
	{
	    struct dirent *dirent;

	    while((dirent = readdir(dir)))
	    {
		if (!strcmp(".", dirent->d_name))
		    continue;

		if (!strncmp("track", dirent->d_name, 5))
		{
		    track_t	*trk;
		    int		 number;
		    int		 vtake;

		    gtk_list_store_append(ls, &row);
		    number = track_char_to_num(dirent->d_name[5]);
		    vtake = dirent->d_name[7] - '0';
		    trk = track_new(number, vtake);
		    prj->track[trk->number][trk->vtake] = trk;
		    track_update(prj, trk, &row);
		}
	    }
	    closedir(dir);
	}
	g_free(dirname); 
    }

    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(ls), TRACK_COL_NUMBER_TEXT,
					 GTK_SORT_ASCENDING);
}

/* Project handling */

project_t *project_new(gint indx)
{
    project_t *proj;

    proj = malloc(sizeof(*proj));
    memset(proj, 0, sizeof(*proj));
    proj->name[0] = '\0';
    proj->number = indx;
    sprintf(proj->filename, "proj%3.3d", indx);
    proj->on_card = FALSE;
    return proj;
}

static void project_get_name(project_t *prj)
{
    gchar	*zif;
    FILE	*F;

    zif = g_build_filename(mrs8path, prj->filename, "prjdata.zif", NULL);
    F = fopen(zif, "rb");
    if (F == NULL)
    {
	prj->name[0] = '\0';
	prj->on_card = FALSE;
	goto out;
    }

    fseek(F, 76, SEEK_SET);
    fread(prj->name, 8, 1, F);
    prj->name[sizeof(prj->name)-1] = '\0';
    fclose(F);
out:
    g_free(zif);
}

static void project_update(project_t *prj, GtkTreeIter *row)
{
    GtkListStore	*ls;
    char		 buf[10];

    ls = GTK_LIST_STORE(gtk_tree_view_get_model(gui_get(WinMain, "ProjectTreeView")));

    project_get_name(prj);
    sprintf(buf, "%3.3d", prj->number);

    gtk_list_store_set(ls, row, PROJECT_COL_DATA, prj, -1);
    gtk_list_store_set(ls, row, PROJECT_COL_NUMBER, prj->number, -1);
    gtk_list_store_set(ls, row, PROJECT_COL_NUMBER_TEXT, buf, -1);
    gtk_list_store_set(ls, row, PROJECT_COL_NAME, prj->name, -1);
}

project_t *project_selected_get(void)
{
    GtkTreeModel	*model;
    GtkTreeIter		 row1;
    gint		 indx;

    if (!gtk_tree_selection_get_selected(
	   gtk_tree_view_get_selection(gui_get(WinMain, "ProjectTreeView")), &model, &row1))
	return NULL;
    gtk_tree_model_get(model, &row1, PROJECT_COL_NUMBER, &indx, -1);
    return project[indx];
}

void project_select(project_t *prj)
{
    GtkTreeView		*tv;
    GtkTreePath		*path;
    GtkTreeModel	*model;
    GtkTreeSelection	*sel;

    tv = gui_get(WinMain, "ProjectTreeView");
    model = gtk_tree_view_get_model(tv);
    sel = gtk_tree_view_get_selection(tv);
    path = gtk_tree_path_new_from_indices(prj->number, -1);
    gtk_tree_selection_select_path(sel, path);

    //gtk_tree_model_get_iter(model, &row, path);
    //gtk_tree_view_set_cursor(tv, path, NULL, FALSE);
    gtk_tree_path_free(path);
}

void project_selected(void)
{
    project_t	*prj;
    char	 name[9];

    prj = project_selected_get();
    track_list_populate(prj);
    ignore_changed = TRUE;
    if (prj == NULL)
    {
	gtk_entry_set_text(GTK_ENTRY(gui_get(WinMain, "ProjectName")), "");
    }
    else
    {
	sprintf(name, "%-8.8s", prj->name);
	gtk_entry_set_text(GTK_ENTRY(gui_get(WinMain, "ProjectName")), name);
    }
    ignore_changed = FALSE;
}

void project_move(gint direction)
{
    project_t		*proj1;
    project_t		*proj2;
    gint		 proj2indx;
    gchar		*old;
    gchar		*new;
    GtkListStore	*ls;
    GtkTreePath		*path;
    GtkTreeIter		 row1;
    GtkTreeIter		 row2;

    proj1 = project_selected_get();
    if (!proj1)
	return;

    proj2indx = proj1->number + direction;
    if (proj2indx < 0 || proj2indx >= MAX_PROJECTS)
	return;
    proj2 = project[proj2indx];

    /* Get rows of GUI */

    ls = GTK_LIST_STORE(gtk_tree_view_get_model(gui_get(WinMain, "ProjectTreeView")));

    path = gtk_tree_path_new_from_indices(proj1->number, -1);
    gtk_tree_model_get_iter(GTK_TREE_MODEL(ls), &row1, path);
    gtk_tree_path_free(path);

    path = gtk_tree_path_new_from_indices(proj2->number, -1);
    gtk_tree_model_get_iter(GTK_TREE_MODEL(ls), &row2, path);
    gtk_tree_path_free(path);

    /* Swap projects in filesystem */

    old = g_build_filename(mrs8path, proj2->filename, NULL);
    new = g_build_filename(mrs8path, "projTEMP", NULL);
    rename(old, new);
    g_free(old);
    g_free(new);

    old = g_build_filename(mrs8path, proj1->filename, NULL);
    new = g_build_filename(mrs8path, proj2->filename, NULL);
    rename(old, new);
    g_free(old);
    g_free(new);

    old = g_build_filename(mrs8path, "projTEMP", NULL);
    new = g_build_filename(mrs8path, proj1->filename, NULL);
    rename(old, new);
    g_free(old);
    g_free(new);

    /* Do the swap */
    {
	int  tmpnum;
	char tmpname[NAME_MAX];

	/* In the array */
	project[proj1->number] = proj2;
	project[proj2->number] = proj1;

	/* Updates names/numbers in the project itself */
	tmpnum = proj1->number;
	strcpy(tmpname, proj1->filename);
	proj1->number = proj2->number;
	strcpy(proj1->filename, proj2->filename);
	proj2->number = tmpnum;
	strcpy(proj2->filename, tmpname);
    }

    /* Update GUI */
    project_update(proj1, &row2);
    project_update(proj2, &row1);
    project_select(proj1);
    return;
}



void project_list_populate(void)
{
    int			 i;
    GtkListStore	*ls;
    GtkTreeView		*tv;
    GtkTreeIter		 row;
    GtkTreeViewColumn	*col;
    GtkTreeSelection	*sel;
    GtkCellRenderer	*rend;

    tv = gui_get(WinMain, "ProjectTreeView");
    ls = GTK_LIST_STORE(gtk_tree_view_get_model(tv));
    if (ls == NULL)
    {
	ls = gtk_list_store_new(PROJECT_COLS, G_TYPE_POINTER, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING);
	gtk_tree_view_set_model(tv, GTK_TREE_MODEL(ls));

	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(ls), PROJECT_COL_NUMBER,
					sort_num_func, GINT_TO_POINTER(PROJECT_COL_NUMBER), NULL);


	rend = gtk_cell_renderer_text_new();
	col  = gtk_tree_view_column_new_with_attributes("Number", rend, "text",
							PROJECT_COL_NUMBER_TEXT, NULL);
	gtk_tree_view_append_column(tv, col);

	col  = gtk_tree_view_column_new_with_attributes("Name", rend, "text",
							PROJECT_COL_NAME, NULL);
	gtk_tree_view_append_column(tv, col);

	sel = gtk_tree_view_get_selection(tv);
	g_signal_connect(G_OBJECT(sel), "changed",
			 G_CALLBACK(project_selected), NULL);
    }

    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(ls),
					 GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID,
					 GTK_SORT_ASCENDING);

    gtk_list_store_clear(ls);

    for (i = 0; i < MAX_PROJECTS; i++)
    {
	gtk_list_store_append(ls, &row);
	project_update(project[i], &row);
    }

    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(ls), PROJECT_COL_NUMBER,
					 GTK_SORT_ASCENDING);
}





gint on_quit(GtkMenuItem *mi, gpointer data)
{
    //if (!lose_changes())
//	return TRUE;

    gtk_main_quit();
    return FALSE;
}


static void open_dir(void)
{
    DIR			*dir;

    dir = opendir(mrs8path);
    if (dir)
    {
	struct dirent *dirent;

	while((dirent = readdir(dir)))
	{
	    if (!strcmp(".", dirent->d_name))
		continue;

	    if (!strncmp("proj", dirent->d_name, 4))
	    {
		int number;
		project_t *prj;

		number = atoi(&dirent->d_name[4]);
		if (number >= 0 && number < MAX_PROJECTS)
		{
		    prj = project[number];
		    prj->on_card = TRUE;
		}
		else
		   gui_modal_ask (FALSE, "Error", "Ok", "",
				  "Invalid project number %d on card", number);
	    }
	}
	closedir(dir);
    }
    project_list_populate();
    gui_window_title(WinMain, PROGRAM " " VERSION " - %s", mrs8path);
}


int main(int argc, char *argv[])
{
    int i;

    gtk_set_locale();
    gtk_init(&argc, &argv);

    WinMain = create_MainWindow();
    gtk_widget_show(WinMain);

    for (i = 0; i < MAX_PROJECTS; i++)
	project[i] = project_new(i);

    /* open file specified on cmd line */
    if (argc > 1)
    {
	strcpy(mrs8path, argv[1]);
	open_dir();
    }
    else
	gui_window_title(WinMain, PROGRAM " "  VERSION);

    gtk_main();
    return 0;
}


void on_open(GtkMenuItem *mi, gpointer data)
{

    on_close(NULL, NULL);
    if (gui_modal_filename_ask("Open MRS8 directory", mrs8path))
	open_dir();
}



void on_close(GtkMenuItem *mi, gpointer data)
{
#if 0
    if (!pts)
	return;

    if (!lose_changes())
	return;

    PS02_pts_close(pts);
    if (pch_file[0])
	PS02_pch_close(pch_file[0]);
    if (pch_file[1])
	PS02_pch_close(pch_file[1]);
    pts = NULL;
    pch_file[0] = pch_file[1] = NULL;
    song_list_populate();
    song_selected();
    pattern_list_populate();
    patch_list_populate(0);
    patch_list_populate(1);
    window_title(PROGRAM " "  VERSION);
#endif
}


void on_about(GtkMenuItem *mi, gpointer data)
{
#if 0
    static const char gpl_terms[] = 
	"This program is free software; you can redistribute it and/or modify\n"
	"it under the terms of the GNU General Public License version 2 as \n"
	"published by the Free Software Foundation.\n \n"
	"This program is distributed in the hope that it will be useful, but\n"
	"WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
	"See the GNU General Public License for more details.";

    char buf[1024];
    sprintf(buf, 
	PROGRAM " " VERSION "\n"
	"Copyright (C) 2003-2005, Dwight Engen <dengen40@yahoo.com>\n\n"
	"%s", gpl_terms);

    modal_ask(FALSE, "About", "Ok", NULL, buf);
#endif
}


void
on_songname_changed                    (GtkEditable     *editable,
                                        gpointer         user_data)
{

}


void
on_project_up                          (GtkButton       *button,
                                        gpointer         user_data)
{
    project_move(-1);
}


void
on_project_down                        (GtkButton       *button,
                                        gpointer         user_data)
{
    project_move(1);
}


void
on_project_copy                        (GtkButton       *button,
                                        gpointer         user_data)
{

}

static void callback(int percent)
{
    GtkWidget *ProgressBar;

    ProgressBar = gui_get(WinProgress, "ProgressBar");
    gtk_progress_set_value(GTK_PROGRESS(ProgressBar), (gdouble)percent);
    gui_sync();
}

void
on_track2wave                          (GtkButton       *button,
                                        gpointer         user_data)
{
    project_t	*prj;
    track_t	*trk[MAX_TRACKS*MAX_VTAKES];
    int		 tracks;
    int		 i;
    char	 outpath[PATH_MAX + 1];
    FILE	*in[2];
    FILE	*out;

    prj = project_selected_get();
    if (prj == NULL)
	return;

    memset(trk, 0, sizeof(trk));
    tracks = tracks_selected(trk);

    if (tracks > 2 || tracks <= 0)
    {
	gui_modal_ask (FALSE, "Error", "Ok", "",
		       "Invalid number of tracks selected (%d).\n"
		       "You must selected only 1 or 2 tracks", tracks);
	return;
    }

    if (!gui_modal_filename_ask("Enter output .wav filename", outpath))
	return;

    for(i = 0; i < tracks; i++)
    {
	gchar		*zaf;

	zaf = g_build_filename(mrs8path, prj->filename, trk[i]->filename, NULL);
	if ((in[i] = fopen(zaf, "rb")) == NULL)
	{
	    gui_modal_ask (FALSE, "Error opening file", "Ok", "",
			   "Unable to open input file %s (%s)\n", trk[i]->filename,
			   strerror(errno));
	    return;
	}
	g_free(zaf);
    }

    if ((out = fopen(outpath, "wb")) == NULL)
    {
	gui_modal_ask (FALSE, "Error opening file", "Ok", "",
		       "Unable to open output file %s (%s)\n", outpath,
		       strerror(errno));
	return;
    }

    WinProgress = create_ProgressWindow();
    gtk_widget_show(WinProgress);

    gui_sync();
    auds2wav(in, out, tracks, callback);
    gtk_widget_destroy(WinProgress);

    for(i = 0; i < tracks; i++)
	fclose(in[i]);
    fclose(out);
}

void on_track_delete (GtkButton *button, gpointer user_data)
{
    project_t	*prj;
    track_t	*trk[MAX_TRACKS*MAX_VTAKES];
    int		 tracks;
    int		 i;

    prj = project_selected_get();
    if (prj == NULL)
	return;

    printf("Hi\n");
    fflush(stdout);
    memset(trk, 0, sizeof(trk));
    tracks = tracks_selected(trk);

    for(i = 0; i < tracks; i++)
    {
	gchar		*zaf;
	zaf = g_build_filename(mrs8path, prj->filename, trk[i]->filename, NULL);
	if (unlink(zaf) < 0)
	    gui_modal_ask (FALSE, "Error unlinking file", "Ok", "",
			   "Unable to unlink file %s (%s)\n", zaf,
			   strerror(errno));
	g_free(zaf);
    }

    track_list_populate(prj);
}
