/*
 * libzoom - Manage the data of a Zoom PS-02 or PS-04.
 *
 * Copyright (C) 2003-2004 Dwight Engen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: $
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "libzoom.h"

#define DEBUG 0		/* 0-4 less to more verbose */
#define PATTERNS_RAW_SIZE (MAX_MEASURES * (MAX_BEATS_PER_MEASURE + 1))

static int Ftruncate(int fd, off_t length);

#ifdef DEBUG_FIELDOFFSETS

#define GET_FIELD(x)						\
{								\
long pos = ftell(in);						\
printf("GET_FIELD Offset:0x%x sz:0x%x\n", pos, sizeof(x));	\
fread(&(x), sizeof(x), 1, in);					\
}

#define PUT_FIELD(x)						\
{								\
long pos = ftell(out);						\
printf("PUT_FIELD Offset:0x%x sz:0x%x\n", pos, sizeof(x));	\
fwrite(&(x), sizeof(x), 1, out);				\
}

#define PUT_ZEROS(n)						\
{								\
int i;								\
long pos = ftell(out);						\
printf("PUT_ZEROS Offset:0x%x sz:%d\n", pos, (n));		\
for (i = 0; i < n; i++)						\
  fwrite("\0", 1, 1, out);					\
}

#else

#define GET_FIELD(x) fread(&x, sizeof(x), 1, in);
#define PUT_FIELD(x) fwrite(&x, sizeof(x), 1, out);
#define PUT_ZEROS(n) { int i; for(i = 0; i < n; i++) fwrite("\0", 1, 1, out); }

#endif

const char note_root_name[][4] = {"C", "C#", "D", "D#", "E", "F", "F#",
				  "G", "G#", "A", "A#", "B"};
const char note_tone_name[][4] = {"   ", "MIN", "7TH", "MI7", "MA7", "AUG",
				  "DIM", "7S4", "SU4", "M7S", "MI6", "MA6",
				  "MI9", "MA9", "MM7", "NON"};

#if 0 /* FIXME: remove? */
static const char *note_name(int note)
{
    static const char name[][40] = {
	"Standard kick",
	"Studio snare",
	"Closed Hihat",
	"Open Hihat",
	"High Tom",
	"Low Tom",
	"Ride Cymbal",
	"Crash cymbal",
	"Stick",
	"Cowbell",
	"Conga Hi",
	"Conga slap",
	"Bongo Lo",
	"Bongo Hi",
	"Room kick",
	"Room snare",
	"Closed Hihat",
	"Jazz kick",
	"Jazz snare",
	"Hiphop kick",
	"Hiphop snare",
	"Hiphop closed hat",
	"Hiphop open hat",
	"Hiphop High Tom",
	"Hiphop Low Tom",
	"Low timbale",
	"High Timbale",
	"High Agogo",
	"Clave",
	"Power kick",
	"Power snare",
	"Funk snare",
	"Met click",
	"Met bell",
	"Low Agogo",
	"EMPTY",
	"EMPTY",
	"EMPTY",
	"Mid Tom",
	"Hiphop Mid Tom",
    };
    if (note <= 40)
	return name[note];
    else
	return "Unknown";
}
#endif


const char *PS0x_midi_drum_name(int midi_note)
{
    static const char name[][40] = {
	"Sticks",		/* 31 */
	"Square click",		/* 32 */
	"Met click",		/* 33 */
	"Met Bell",		/* 34 */
	"Kick 2",		/* 35 */
	"Kick 1",		/* 36 */
	"Sidestick",		/* 37 */
	"Snare 1",		/* 38 */
	"Claps",		/* 39 */
	"Snare 2",		/* 40 */
	"Lo Tom 2",		/* 41 */
	"Closed hat",		/* 42 */
	"Lo Tom1",		/* 43 */
	"Pedal Hat",		/* 44 */
	"Mid Tom 2",		/* 45 */
	"Open hat",		/* 46 */
	"Mid Tom 1",		/* 47 */
	"Hi  Tom 2",		/* 48 */
	"Crash cymbal 1",	/* 49 */
	"Hi Tom 1",		/* 50 */
	"Ride cymbal 1",	/* 51 */
	"Chinese cymbal",	/* 52 */
	"Ride bell",		/* 53 */
	"Tambourine",		/* 54 */
	"Splash cymbal",	/* 55 */
	"Cow bell",		/* 56 */
	"Crash cymbal 2",	/* 57 */
	"Quijada",		/* 58 */
	"Ride cymbal 2",	/* 59 */
	"Hi Bongo",		/* 60 */
	"Lo Bongo",		/* 61 */
	"Conga slap",		/* 62 */
	"Conga Hi",		/* 63 */
	"Conga Lo",		/* 64 */
	"Timbale Hi",		/* 65 */
	"Timbale Lo",		/* 66 */
	"Agogo Hi",		/* 67 */
	"Agogo Lo",		/* 68 */
	"Cabasa",		/* 69 */
	"Maracas",		/* 70 */
	"Whistle short",	/* 71 */
	"Whistle Long",		/* 72 */
	"Guiro short",		/* 73 */
	"Guiro long",		/* 74 */
	"Claves",		/* 75 */
	"Woodblock Hi",		/* 76 */
	"Woodblock Lo",		/* 77 */
	"Mute Quica",		/* 78 */
	"Open Quica",		/* 79 */
	"Mute triangle",	/* 80 */
	"Open triangle",	/* 81 */
	"Shaker",		/* 82 */
	"Sleighbells",		/* 83 */
	"Castanets",		/* 84 */
	"Mute Surdo",		/* 85 */
	"Open Surdo",		/* 86 */
	"Unknown",		/* 87 */
	"Applause"		/* 88 */
    };
    if (midi_note >= 31 && midi_note <= 88)
	return name[midi_note-31];
    else
	return "Unknown";
}


const char *PS0x_midi_note_name(int midi_note)
{
    int octave;
    static const char name[][40] = {
	"C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"
    };
    static char gen_name[10];
    
    octave = midi_note / 12 - 1;
    sprintf(gen_name, "%s%d", name[midi_note % 12], octave);
    return gen_name;
}



static void PS02_pattern_unpack(ps02_pattern_t *ptn)
{
    int i;
    int ppq_add = 0;
    int ppq_total = 0;
    int drum_bank_base = 36;
    ptn_event_t *bass_on_ev = NULL;

    /* Bass bank note to MIDI note maps */
    int  bass_map[] = {21,23,24,27,29,33,33,35,36,39,29,33,33,35,	/* 0x0e */
		       22,23,26,28,31,32,34,35,38,28,31,32,34,35,	/* 0x0e, 0x80 */

		       33,35,36,39,41,45,45,35,36,39,41,45,45,47,	/* 0x1e */
		       34,35,38,28,31,32,34,35,38,40,43,44,46,47,

		       33,35,36,39,41,45,45,47,48,51,41,45,45,47,	/* 0x2e */
		       34,35,38,40,43,44,46,47,50,52,43,44,46,47};
    int *bass_bank = &bass_map[14 * 2];

    ptn->event = malloc(MAX_EVENTS_PER_PATTERN * sizeof(ptn_event_t));
    ptn->events = 0;

#   if DEBUG > 3
    Log(4, "BYTES Event PPQDelta Type      Vel MIDI Name\n");
#   endif

    for(i = 0; i < ptn->raw_bytes; i += 2)
    {
	ptn_event_t *ev = &ptn->event[ptn->events];

	unsigned char b1;
	unsigned char b2;


	b1 = ptn->raw[i];
	b2 = ptn->raw[i+1];

	ev->ppq_delta = (b1 & 0x1f) * PS02_PPQ_FACTOR + ppq_add;
	ev->ppq_duration = 0;
	if (bass_on_ev)
	    bass_on_ev->ppq_duration += ev->ppq_delta;

	switch(b1 & 0x60)
	{
	/* Drum event */
	case 0x60:
	    ppq_total += ev->ppq_delta;
	    ppq_add = 0;
	    switch(b2)
	    {
	    /* Bank switch events */
	    case 0x0e:
	    case 0x1e:
	    case 0x2e:
	    case 0x3e:
		ev->type = PTN_EVENT_DRUM_BANK;
		ev->midi_note = b2;
		ev->velocity15 = 0;
	    default:
		break;
	    }

	    switch(b2)
	    {
	    /* Bank switch events */
	    case 0x0e:
		drum_bank_base = 24;
		break;
	    case 0x1e:
		drum_bank_base = 36;
		break;
	    case 0x2e:
		drum_bank_base = 48;
		break;
	    case 0x3e:
		drum_bank_base = 60;
		break;
	    case 0x0f:
		ev->type = PTN_EVENT_END;
		ev->midi_note = 0x0f;
		ev->velocity15 = 0;
		break;

	    /* Drum hit event */
	    default:
		ev->type = PTN_EVENT_DRUM;
		ev->midi_note = drum_bank_base + (((b2 & 0x0f) << 1) | (b1 >> 7));
		ev->velocity15 = (b2 & 0xf0) >> 4;
	    }
	    break;

	/* Bass event */
	case 0x00:
	    ppq_total += ev->ppq_delta;
	    ppq_add = 0;
	    switch(b2)
	    {
	    /* Bank switch events */
	    case 0x0e:
	    case 0x1e:
	    case 0x2e:
	    case 0x3e:
		ev->type = PTN_EVENT_BASS_BANK;
		ev->midi_note = b2;
		ev->velocity15 = 0;
	    default:
		break;
	    }

	    switch(b2)
	    {
	    /* Bank switch events */
	    case 0x0e:
		bass_bank = &bass_map[0];
		break;
	    case 0x1e:
		bass_bank = &bass_map[14 * 2];
		break;
	    case 0x2e:
		bass_bank = &bass_map[14 * 4];
		break;
	    case 0x0f:
		ev->type = PTN_EVENT_END;
		ev->midi_note = 0x0f;
		ev->velocity15 = 0;
		break;

	    /* Bass note event */
	    default:
		Assert((b2 & 0xf) <= 0x0d, "WARNING: note out of range!\n");

		ev->type = PTN_EVENT_BASS;
		if (b1 & 0x80)
		    ev->midi_note = bass_bank[(b2 & 0x0f) + 14];
		else
		{
		    ev->midi_note = bass_bank[b2 & 0x0f];

		    if (ptn->bass_octave < 0x9f)
			ev->midi_note += 12;
		}

		ev->velocity15 = (b2 & 0xf0) >> 4;

		/* Remeber the event where the bass went on so we can fill in
		 * it's ppq duration when we see it go off again. This is
		 * needed since the PS04 wants to know the bass note's
		 * duration at the time of sounding instead since it doesn't
		 * get an off event.
		 */
		if (ev->velocity15 > 0)
		    bass_on_ev = ev;
		else
		{
		    if (bass_on_ev)
			bass_on_ev = NULL;
		}
	    }
	    break;

	/* Delay event */
	default:
	    ppq_add += (((b2 << 3) | (b1 >> 5)) + 0x1f) * PS02_PPQ_FACTOR;
	    ev->ppq_delta = 0;
	    switch(b2)
	    {
	    case 0x0f:
		ev->type = PTN_EVENT_END;
		ev->midi_note = 0x0f;
		ev->velocity15 = 0;
		break;
	    default:
		ev->type = PTN_EVENT_PPQ;
		ev->midi_note = ppq_add;
		ev->velocity15 = 0;
	    }
	    break;
	}


#	if DEBUG > 3	/* Dump translated event */
	Log(4, "%2.2x %2.2x ", b1, b2);
	printf("%5d %8d ", ptn->events+1, ev->ppq_delta);
	switch(ev->type)
	{
	case PTN_EVENT_DRUM:		printf("Drum      "); break;
	case PTN_EVENT_DRUM_BANK:	printf("Drum bank "); break;
	case PTN_EVENT_BASS:		printf("Bass      "); break;
	case PTN_EVENT_BASS_BANK:	printf("Bass bank "); break;
	case PTN_EVENT_PPQ:		printf("PPQ delay "); break;
	case PTN_EVENT_END:		printf("PTN END   "); break;
	}

	printf("%3d %4d ", ev->velocity15, ev->midi_note);

	switch(ev->type)
	{
	case PTN_EVENT_DRUM:
	    printf("%s\n", PS0x_midi_drum_name(ev->midi_note));
	    break;
	case PTN_EVENT_BASS:
	    printf("%s %s\n", PS0x_midi_note_name(ev->midi_note), ev->velocity15 ? "" : "off");
	    break;
	default:
	    printf("\n");
	    break;
	}
#	endif

	/* Let a PPQ event get overridden by the next event */
	if (ev->type != PTN_EVENT_PPQ)
	    ptn->events++;
    }
    Log(2, "PPQ Total:  %8d delay:%d\n", ppq_total,
	ptn->bars * ptn->beats * PS02_PPQ * PS02_PPQ_FACTOR - ppq_total);
    Log(2, "Rhythm events:%d\n", ptn->events);
}



ps02_pattern_t *PS02_pattern_read(FILE *in)
{
    ps02_pattern_t	*ptn;

    ptn = malloc(sizeof(ps02_pattern_t));

    GET_FIELD(ptn->name);
    GET_FIELD(ptn->name_pad);
    GET_FIELD(ptn->bpm);
    GET_FIELD(ptn->beats);
    GET_FIELD(ptn->bars);
    GET_FIELD(ptn->unknown2);
    GET_FIELD(ptn->bass_octave);
    GET_FIELD(ptn->unknown3);
    GET_FIELD(ptn->raw_bytes);
    Log(2, "Rhythm pattern:%6.6s beats:%d bars:%d bytes:%d oct:%x\n",
	ptn->name, ptn->beats, ptn->bars, ptn->raw_bytes, ptn->bass_octave);

    ptn->raw = malloc(ptn->raw_bytes);
    fread(ptn->raw, ptn->raw_bytes, 1, in);
    PS02_pattern_unpack(ptn);
    return ptn;
}



void PS02_pattern_write(FILE *out, ps02_pattern_t *ptn)
{
    Log(2, "Rhythm pattern:%6.6s bytes:%d\n", ptn->name, ptn->raw_bytes);

    PUT_FIELD(ptn->name);
    PUT_FIELD(ptn->name_pad);
    PUT_FIELD(ptn->bpm);
    PUT_FIELD(ptn->beats);
    PUT_FIELD(ptn->bars);
    PUT_FIELD(ptn->unknown2);
    PUT_FIELD(ptn->bass_octave);
    PUT_FIELD(ptn->unknown3);
    PUT_FIELD(ptn->raw_bytes);
    fwrite(ptn->raw, ptn->raw_bytes, 1, out);
}



static void PS02_song_pack(ps02_song_t *song)
{
    int			 i;
    unsigned char	*ch;

    if (song->measures == 0)
	return;

    Log(2, "measures:%d max bytes:%d\n", song->measures,
	song->measures * (MAX_BEATS_PER_MEASURE + 1));

    ch = song->raw;

    *ch++ = 0x00;
    for (i = 0; i < song->measures; i++)
    {
	int beats = 0;
	int j;
	int last_note_beat;

	*ch++ = song->measure[i].pattern;

#if 0 //global pts dependency
	Log(2, "Measure:%2d Pattern:%6.6s (0x%2.2x) Beats:%d\n", i,
	    pts->rhy_pattern[song->measure[i].pattern]->name,
	    song->measure[i].pattern,
	    song->measure[i].beats);
#endif

	last_note_beat = 0;
	for (j = 0; j < song->measure[i].beats; j++)
	{
	    unsigned char note = song->measure[i].note[j];

	    if (note != 0xff)
	    {
		Log(2, "Beat:%2d Note:%s (0x%2.2x) LastNoteBeat:%d\n",
		    j, note_root_name[note & 0x0f], note, last_note_beat);

		beats += (j - last_note_beat);
		*ch++ = 0x20 | (j - last_note_beat);
		last_note_beat = j;
		*ch++ = note;
	    }
	}

	/* status byte */
	*ch++ = song->measure[i].beats - beats;
    }
    *(ch-1) |= 0xE0;

    *ch++ = 0xff;
    song->patterns_len = ch-song->raw;

    if (song->patterns_len >= PATTERNS_RAW_SIZE)
    {
	printf("Patterns too big:%d\n", song->patterns_len);
	exit(10);
    }

    Log(2, "Bytes of pattern data:%d (0x%x)\n",
	song->patterns_len, song->patterns_len);
}



static void PS02_song_unpack(ps02_song_t *song)
{
    int			 measure;
    unsigned char	 beat;
    unsigned char	*b;

    song->measure = malloc(sizeof(ps02_measure_t) * MAX_MEASURES);
    memset(song->measure, 0xff, sizeof(ps02_measure_t) * MAX_MEASURES);

    measure = 0;
    beat = 0;
    for(b = song->raw+1; b < song->raw+song->patterns_len; )
    {
	unsigned char	 note;
	unsigned char	 status;

	/* Get measure's pattern */
	song->measure[measure].pattern = *b++;

	/* Get all chord changes */
	while ((*b & 0xF0) == 0x20)
	{
	    beat += *b++ & 0x0f;
	    note  = *b++;
	    if (beat >= MAX_BEATS_PER_MEASURE)
	    {
		printf("too many beats!\n");
		exit(1);
	    }
	    song->measure[measure].note[beat] = note;
	}
	status = *b++;
	beat += status & 0x0f;
	song->measure[measure].beats = beat;

#if 0 //global pts dependency
	Log(2, "Measure %2d Pattern:%6.6s (0x%2.2x) Beats:%d stat:0x%2.2x\n", measure,
	    pts->rhy_pattern[song->measure[measure].pattern]->name,
	    song->measure[measure].pattern,
	    song->measure[measure].beats,
	    status);
#endif

	measure++;
	beat = 0;

	if ((status & 0xe0) == 0xe0)
	{
	    song->measures = measure;
	    break;
	}
    }
}



ps02_song_t *PS02_song_copy(ps02_song_t *origsong)
{
    ps02_song_t	*song;

    song = malloc(sizeof(*song));
    memcpy(song, origsong, sizeof(*song));

    song->raw = malloc(PATTERNS_RAW_SIZE);
    memset(song->raw, 0, PATTERNS_RAW_SIZE);
    memcpy(song->raw, origsong->raw, song->patterns_len);

    PS02_song_unpack(song);
    return song;
}



void PS02_song_double(ps02_song_t *song)
{
    memcpy(&song->measure[song->measures], &song->measure[0],
	   song->measures * sizeof(ps02_measure_t));
    song->measures *= 2;
}



void PS02_song_delete_measures(ps02_song_t *song, int start, int end)
{
    int measures = end - start + 1;

    memmove(&song->measure[start], &song->measure[end+1],
	    (song->measures-end) * sizeof(ps02_measure_t));

    song->measures -= measures;
}



void PS02_song_copy_measures(ps02_song_t *song, int start, int end, int dest)
{
    int measures = end - start + 1;
    int size = measures * sizeof(ps02_measure_t);
    char *tmpbuf;

    tmpbuf = malloc(size);
    memcpy(tmpbuf, &song->measure[start], size);

    /* Slide dest up size measures */
    memmove(&song->measure[dest+measures], &song->measure[dest],
	    (song->measures-dest) * sizeof(ps02_measure_t));

    memcpy(&song->measure[dest], tmpbuf, size);
    song->measures += measures;
    free(tmpbuf);
}



void PS02_song_free(ps02_song_t *song)
{
    free(song->measure);
    free(song->raw);
    free(song);
}



ps02_song_t *PS02_song_read(FILE *in)
{
    ps02_song_t	*song;

    song = malloc(sizeof(ps02_song_t));

    GET_FIELD(song->name);
    GET_FIELD(song->drum_volume);
    GET_FIELD(song->bass_volume);
    GET_FIELD(song->input_volume);
    GET_FIELD(song->pad1);
    GET_FIELD(song->patterns_len);
    GET_FIELD(song->tempo);
    GET_FIELD(song->repeat);
    GET_FIELD(song->drum_kit);
    GET_FIELD(song->bass_kit);

    song->raw = malloc(PATTERNS_RAW_SIZE);
    memset(song->raw, 0, PATTERNS_RAW_SIZE);
    fread(song->raw, song->patterns_len, 1, in);
    PS02_song_unpack(song);
    return song;
}



void PS02_song_write(FILE *out, ps02_song_t *song)
{
    PS02_song_pack(song);

    PUT_FIELD(song->name);
    PUT_FIELD(song->drum_volume);
    PUT_FIELD(song->bass_volume);
    PUT_FIELD(song->input_volume);
    PUT_FIELD(song->pad1);
    PUT_FIELD(song->patterns_len);
    PUT_FIELD(song->tempo);
    PUT_FIELD(song->repeat);
    PUT_FIELD(song->drum_kit);
    PUT_FIELD(song->bass_kit);

    fwrite(song->raw, song->patterns_len, 1, out);
}

void PS02_patch_read(FILE *in, patch_t *patch)
{
#define ZWID 6
    unsigned char	 tmp[7] = {0,0,0,0,0,0,0};
    char		*name;
    static const char PatchCharSet[64] =
    {
	    ' ', '\'', '<', '>', '*', '+', '-', '/', '0', '1',
	    '2', '3',  '4', '5', '6', '7', '8', '9', '[', '=',
	    ']', '@',  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
	    'I', 'J',  'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R',
	    'S', 'T',  'U', 'V', 'W', 'X', 'Y', 'Z', '_', ' ',
	    '0', '0',  '0', '0', '0', '0', '0', '0', '0', '0',
	    '0', '0',  '0', '0'
    };

    fread(patch->data, 1, sizeof(patch->data), in);

    /* Unpack name */
    memset(patch->name, 0, sizeof(patch->name));
    name = patch->name;
    memcpy(&tmp, &patch->data[11], 6);

    {
	int j;
	int ish = 5;
	int idx = 0;

	for(j = 0; j < 6; j++)
	{
	    name[j] = PatchCharSet[(*((short*)(&tmp[idx])) >> ish) & (char)((0x1<<ZWID)-1)];
	    ish -= (8*sizeof(char) - ZWID);
	    if(0 > ish)
	    {
		ish += 8*sizeof(char);
		idx--;
	    }
	    idx++;
	}
    }
#undef ZWID
}

void PS02_patch_write(FILE *out, patch_t *patch)
{
	fwrite(patch->data, 1, sizeof(patch->data), out);
}

void PS02_pch_read(FILE *in, ps02_pch_t *pch)
{
    int			 i;

    fseek(in, 0, SEEK_SET);
    for(i=0; i < NUM_PATCHES; i++)
	PS02_patch_read(in, &pch->pch[i]);
}

void PS02_pch_write(FILE *out, ps02_pch_t *pch)
{
    int			 i;

    fseek(out, 0, SEEK_SET);
    for(i=0; i < NUM_PATCHES; i++)
	PS02_patch_write(out, &pch->pch[i]);
}

ps02_pch_t *PS02_pch_open(const char *file)
{
    ps02_pch_t *newpch;

    newpch = malloc(sizeof(*newpch));
    memset(newpch, 0, sizeof(*newpch));

    newpch->F = fopen(file, "r+b");
    if (newpch->F == NULL)
    {
	free(newpch);
	return NULL;
    }

    PS02_pch_read(newpch->F, newpch);
    return newpch;
}

void PS02_pch_close(ps02_pch_t *pch)
{
    fclose(pch->F);
    free(pch);
}



void PS02_pts_read(ps02_pts_t *pts)
{
    FILE	*in;
    int		 i;

    in = pts->F;

    GET_FIELD(pts->header);
    GET_FIELD(pts->rhy_pattern_cnt);
    Log(2, "Rhythm patterns: %d\n", pts->rhy_pattern_cnt);
    for(i=0; i < pts->rhy_pattern_cnt; i++)
	pts->rhy_pattern[i] = PS02_pattern_read(in);

    pts->rhy_pattern[0xff] = malloc(sizeof(ps02_pattern_t));
    memcpy(pts->rhy_pattern[0xff]->name, "PTN<--", 6);

    GET_FIELD(pts->rhy_grp0_indx);
    GET_FIELD(pts->rhy_grp1_indx);
    GET_FIELD(pts->rhy_grp2_indx);

    /* Read rhythm main group count and names */
    GET_FIELD(pts->rhy_main_group_cnt);
    pts->rhy_main_group_names = malloc(6 * pts->rhy_main_group_cnt);
    fread(pts->rhy_main_group_names, pts->rhy_main_group_cnt, 6, in);

    /* Read rhythm sub group count and names */
    GET_FIELD(pts->rhy_sub_group_cnt);
    pts->rhy_sub_group_names = malloc(6 * pts->rhy_sub_group_cnt);
    fread(pts->rhy_sub_group_names, pts->rhy_sub_group_cnt, 6, in);

    /* Read in all songs */
    for (i=0; i < 100; i++)
    {
	pts->song[i] = PS02_song_read(in);
	Log(2, "%3d %6.6s\n", i+1, pts->song[i]->name);
    }
}



void PS02_pts_write(ps02_pts_t *pts)
{
    FILE *out;
    int	i;

    out = pts->F;
    fseek(out, 0, SEEK_SET);
    PUT_FIELD(pts->header);
    PUT_FIELD(pts->rhy_pattern_cnt);
    Log(2, "Rhythm patterns: %d\n", pts->rhy_pattern_cnt);
    for(i=0; i < pts->rhy_pattern_cnt; i++)
	PS02_pattern_write(out, pts->rhy_pattern[i]);

    PUT_FIELD(pts->rhy_grp0_indx);
    PUT_FIELD(pts->rhy_grp1_indx);
    PUT_FIELD(pts->rhy_grp2_indx);

    /* Write rhythm main group count and names */
    PUT_FIELD(pts->rhy_main_group_cnt);
    fwrite(pts->rhy_main_group_names, pts->rhy_main_group_cnt, 6, out);

    /* Write rhythm sub group count and names */
    PUT_FIELD(pts->rhy_sub_group_cnt);
    fwrite(pts->rhy_sub_group_names, pts->rhy_sub_group_cnt, 6, out);

    /* Write all songs */
    for (i=0; i < 100; i++)
    {
	Log(2, "%3d %6.6s\n", i+1, pts->song[i]->name);
	PS02_song_write(out, pts->song[i]);
    }
}

ps02_pts_t *PS02_pts_open(const char *file)
{
    ps02_pts_t *newpts;

    newpts = malloc(sizeof(*newpts));
    memset(newpts, 0, sizeof(*newpts));

    newpts->F = fopen(file, "r+b");
    if (newpts->F == NULL)
    {
	free(newpts);
	return NULL;
    }

#if 0 //global pts dependency
    pts = newpts;
#   endif

    PS02_pts_read(newpts);
    return newpts;
}

void PS02_pts_close(ps02_pts_t *pts)
{
    fclose(pts->F);
    free(pts);
}

int PS02_pts_modified(ps02_pts_t *pts)
{
    return pts->modified;
}



static int PS04_vel15_to_vel8(int vel15)
{
    /* Armin mapping: int vel8[] = {0,0,0,1,2,2,2,3,4,4,4,5,6,6,6,7}; */
    int vel8[] = {0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7};
    return vel8[vel15];
}

#if 0 /* FIXME: not yet */
static int PS04_vel8_to_vel16(int vel8)
{
    int vel15[] = {0,3,5,7,9,11,13,15};
    return vel15[vel8];
}
#endif

static void PS04_pattern_pack(ps04_pattern_t *ptn)
{
    int i,j;
    int events = 0;
    int ppq_add = 0;
    int ppq_total = 0;

    if (!ptn->sevents)
	return;

    ptn->raw = realloc(ptn->raw, ptn->sevents * 2 * 4);

    Log(3, "Ptn:%6.6s Beats:%d Bars:%d TotPPQ:%d\n", ptn->name_long,
	ptn->beats, ptn->bars, ptn->beats * ptn->bars * PS04_PPQ);

    for (i = 0, j = 0; i < ptn->sevents; i++)
    {
	ptn_event_t *ev = &ptn->event[i];

	switch(ev->type)
	{
	case PTN_EVENT_DRUM:
	  {
	    unsigned char vel8 = PS04_vel15_to_vel8(ev->velocity15);

	    ppq_total += (ev->ppq_delta + ppq_add) * PS04_PPQ_FACTOR;
	    ptn->raw[j++] = (ev->ppq_delta + ppq_add) * PS04_PPQ_FACTOR;
	    ptn->raw[j++] = (ev->midi_note - 12) | ((vel8 & 1) << 7);
	    vel8 >>= 1;
	    ptn->raw[j++] = vel8;
	    ptn->raw[j++] = 0x00;
	    events++;
	    ppq_add = 0;
	    Log(3, "Ev:%3d Drum RAW:%2.2x%2.2x%2.2x%2.2x %s\n", i+1,
		ptn->raw[j-4], ptn->raw[j-3],
		ptn->raw[j-2], ptn->raw[j-1], PS0x_midi_drum_name(ev->midi_note));
	    break;
	  }

	case PTN_EVENT_BASS:
	  {
	    unsigned char vel8 = PS04_vel15_to_vel8(ev->velocity15);

	    if (ev->ppq_duration == 0)	/* Note off */
	    {
		ppq_add += ev->ppq_delta;
		Log(3, "Ev:%3d Bass off ppq:%d\n", i+1, ev->ppq_delta);
		break;
	    }

	    ppq_total += (ev->ppq_delta + ppq_add) * PS04_PPQ_FACTOR;
	    ptn->raw[j++] = (ev->ppq_delta + ppq_add) * PS04_PPQ_FACTOR;
	    if (ev->midi_note < 28)
		ptn->raw[j++] = (ev->midi_note + 0xc0 - 24) | ((vel8 & 1) << 7);
	    else
		ptn->raw[j++] = (ev->midi_note + 0xd0 - 28) | ((vel8 & 1) << 7);

	    vel8 >>= 1;
	    ptn->raw[j++] = (((ev->ppq_duration * PS04_PPQ_FACTOR) & 0x3f) << 2) | (vel8 & 3);
	    ptn->raw[j++] = (ev->ppq_duration * PS04_PPQ_FACTOR) >> 6;
	    events++;
	    ppq_add = 0;
	    Log(3, "Ev:%3d Bass RAW:%2.2x%2.2x%2.2x%2.2x dur:%d\n", i+1,
		ptn->raw[j-4], ptn->raw[j-3],
		ptn->raw[j-2], ptn->raw[j-1], ev->ppq_duration);
	    break;
	  }

	case PTN_EVENT_END:
	    /* Write loop event */
	    Log(3, "Ev:%3d END ppq_delta:%d ppq_fixup:%d\n",
		i+1, ev->ppq_delta, ptn->beats * ptn->bars * PS04_PPQ - ppq_total);

	    ptn->raw[j++] = ptn->beats * ptn->bars * PS04_PPQ - ppq_total;
	    ptn->raw[j++] = 0x3f;
	    ptn->raw[j++] = 0x00;
	    ptn->raw[j++] = 0x00;
	    events++;
	    ppq_add = 0;
	    break;

	case PTN_EVENT_BASS_BANK:
	case PTN_EVENT_DRUM_BANK:
	    Log(3, "Ev:%3d Bass BNK ppq_delta:%d\n", i+1, ev->ppq_delta);
	    ppq_add += ev->ppq_delta;
	    break;

	default:
	    Log(3, "Ev:%3d %d ignored\n", i+1, ev->type);
	    break;
	}
    }

    ptn->rawevents = events;
    ptn->event_start = 0;
}



void PS04_pattern_write(FILE *out, ps04_pattern_t *ptn)
{
    PS04_pattern_pack(ptn);

    PUT_FIELD(ptn->name_long);
    PUT_FIELD(ptn->name);
    PUT_FIELD(ptn->beats);
    PUT_FIELD(ptn->bars);
    PUT_FIELD(ptn->drum_kit);
    PUT_FIELD(ptn->bass_kit);
    PUT_FIELD(ptn->drum_level);
    PUT_FIELD(ptn->bass_level);
    PUT_FIELD(ptn->original_root);
    PUT_FIELD(ptn->original_chord);
    PUT_FIELD(ptn->rawevents);
    PUT_FIELD(ptn->event_start);

    fwrite(ptn->raw, ptn->rawevents * 4, 1, out);
    Log(3, "%-6.6s events:%d\n", ptn->name_long, ptn->rawevents);
}



ps04_pattern_t *PS04_pattern_read(FILE *in)
{
    ps04_pattern_t	*ptn;

    ptn = malloc(sizeof(*ptn));

    GET_FIELD(ptn->name_long);
    GET_FIELD(ptn->name);
    GET_FIELD(ptn->beats);
    GET_FIELD(ptn->bars);
    GET_FIELD(ptn->drum_kit);
    GET_FIELD(ptn->bass_kit);
    GET_FIELD(ptn->drum_level);
    GET_FIELD(ptn->bass_level);
    GET_FIELD(ptn->original_root);
    GET_FIELD(ptn->original_chord);
    GET_FIELD(ptn->rawevents);
    GET_FIELD(ptn->event_start);

    Log(3, "%-6.6s events:%d\n", ptn->name_long, ptn->rawevents);

    ptn->raw = malloc(ptn->rawevents * 4);
    fread(ptn->raw, ptn->rawevents * 4, 1, in);
    ptn->sevents = 0; //FIXME
    //FIXME: PS04_pattern_unpack(ptn);
    return ptn;
}



ps04_pattern_t *PS04_pattern_new(void)
{
    ps04_pattern_t	*ptn;

    ptn = malloc(sizeof(*ptn));
    memset(ptn, 0, sizeof(*ptn));
    strcpy(ptn->name_long, "(empty)");
    return ptn;
}



static ps04_song_event_t *PS04_emit_ppq(ps04_song_event_t *event, int ppq)
{
    unsigned char N,M;

    /* Small PPQ's that fit into a byte go inside the event itself */
    if (ppq < 0xff)
    {
	event->next_ppq = ppq;
	return event+1;
    }

    /* Larger PPQ's need their own event */
    event->next_ppq = 0xff;

    event++;
    M = (ppq & 0x00f0) | 1;
    N = ((ppq & 0xf000) >> 8) | (((ppq & 0x0f00) >> 8) - 1);

    *(uint32_t *)event = N<<8 | M;
    return event+1;
}



void PS04_song_pack(ps04_song_t *song)
{
    int			 i;
    uint8_t		*ch;
    ps04_song_event_t	*ps04_event;
    unsigned int	 rawsize;

    if (song->events <= 0)
	return;

    /* Ensure enough raw_size that every event *could* have
     * a long PPQ
     */
    rawsize = (song->events*2+1) * 4;
    song->raw = realloc(song->raw, rawsize);
    ch = song->raw;

    *ch++ = 0x00;
    ps04_event = (ps04_song_event_t *)ch;
    for (i = 0; i < song->events; i++)
    {
	ps04_song_event_t *next_event;
	ps04_event->type = song->event[i].type;
	ps04_event->data = song->event[i].data;
	next_event = PS04_emit_ppq(ps04_event, song->event[i].next_ppq);
	ps04_event = next_event;
    }
    ch = (uint8_t *)ps04_event;

    memset(ch, 0x00, 3);
    ch+=3;

    /* Calculate the real raw size and number of hardware events */
    rawsize = ch - song->raw;
    song->rawevents = rawsize / 4;
}



static int PS04_parse_ppq(song_event_t *event, ps04_song_event_t *ev04)
{
#define NIBBLE_SWAP(x) 	(((x) & 0x0f) << 4) | (((x) & 0xf0) >> 4);

    unsigned char N,M,S;

    if (ev04->next_ppq == 0xff)
    {
	/* Larger PPQ's need their own event
	 * PPQ Ticks = ((N & 0xf0) << 12) | (nibble swap(M + (N & 0x0f)) << 4)
	 */

	ev04++;
	N = (*(uint32_t *)ev04 & 0xff00) >> 8;
	M = (*(uint32_t *)ev04 & 0x00ff);
	S = M + (N & 0x0f);
	S = NIBBLE_SWAP(S);
	event->next_ppq = ((N & 0xf0) << 12) | S << 4;
	return 1;
    }
    else
    {
	/* Small PPQ's (less than 2 measures) fit inside the event itself */

	Assert(ev04->next_ppq < 0xff, "Saw PPQ:0x%x", ev04->next_ppq);
	event->next_ppq = ev04->next_ppq;
	return 0;
    }
}



void PS04_song_unpack(ps04_song_t *song)
{
    int			 i;
    int			 j;
    void		*ch;
    ps04_song_event_t	*ps04_event;

    if (song->rawevents <= 0)
	return;

    /* Ensure enough event space that every event *could* have
     * a long PPQ
     */
    song->event = realloc(song->event, song->rawevents * sizeof(song_event_t));

    ch = song->raw + 1;
    ps04_event = ch;
    for (i = 0, j = 0; i < song->rawevents; i++, j++)
    {
	int inc;

	song->event[j].type = ps04_event->type;
	song->event[j].data = ps04_event->data;
	inc = PS04_parse_ppq(&song->event[j], ps04_event);

	ps04_event += 1 + inc;
	i += inc;
    }

    /* Calculate the number of events */
    song->events = j - 1;
}



void PS04_song_write(FILE *out, ps04_song_t *song)
{
    PS04_song_pack(song);

    PUT_FIELD(song->name_long);
    PUT_FIELD(song->name);
    PUT_FIELD(song->rawevents);
    PUT_FIELD(song->unknown1);

    if (song->rawevents > 0)
    {
	fwrite(song->raw, song->rawevents * 4, 1, out);
    }
}



ps04_song_t *PS04_song_read(FILE *in)
{
    ps04_song_t	*song;
    unsigned int rawsize;

    song = malloc(sizeof(*song));

    GET_FIELD(song->name_long);
    GET_FIELD(song->name);
    GET_FIELD(song->rawevents);
    if (song->rawevents == 0)
	return song;

    GET_FIELD(song->unknown1);

    rawsize = song->rawevents * 4;
    song->raw = malloc(rawsize);
    memset(song->raw, 0, rawsize);
    fread(song->raw, rawsize, 1, in);
    PS04_song_unpack(song);

    return song;
}



ps04_song_t *PS04_song_new(void)
{
    ps04_song_t	*song;

    song = malloc(sizeof(*song));
    memset(song, 0, sizeof(*song));
    strcpy(song->name_long, "(Empty)");
    return song;
}



void PS04_seq_write(ps04_seq_t *seq)
{
    FILE	*out;
    int		 i;
    long	 pos;
    long	 size;

    out = seq->F;
    fseek(out, 0, SEEK_SET);
    Ftruncate(fileno(out), 0);

    PUT_FIELD(seq->fmt);
    PUT_FIELD(seq->unknown1);
    PUT_FIELD(seq->patterns);
    PUT_FIELD(seq->songs);
    PUT_FIELD(seq->unknown2);

    Log(2, "Rhythm patterns:%d songs:%d\n",
	seq->patterns, seq->songs);

    for(i=0; i < seq->patterns; i++)
	PS04_pattern_write(out, seq->pattern[i]);

    for (i=0; i < seq->songs; i++)
	PS04_song_write(out, seq->song[i]);

    /* File size must end with 00 */
    pos = ftell(out);
    size = (pos + 0xff) & ~0xff;
    PUT_ZEROS(size-pos);
}



void PS04_seq_read(ps04_seq_t *seq)
{
    FILE	*in;
    int		 i;

    in = seq->F;

    GET_FIELD(seq->fmt);
    GET_FIELD(seq->unknown1);
    GET_FIELD(seq->patterns);
    GET_FIELD(seq->songs);
    GET_FIELD(seq->unknown2);

    Log(2, "Rhythm patterns:%d songs:%d\n",
	seq->patterns, seq->songs);

    for(i=0; i < seq->patterns; i++)
	seq->pattern[i] = PS04_pattern_read(in);

    for (i=0; i < seq->songs; i++)
	seq->song[i] = PS04_song_read(in);
}



ps04_seq_t *PS04_seq_new(void)
{
    ps04_seq_t *seq;
    int		i;

    seq = malloc(sizeof(*seq));
    memset(seq, 0, sizeof(*seq));

    strcpy(seq->fmt, "ZOOM D120SEQ");
    seq->patterns = 511;
    seq->songs = 10;
    for(i=0; i < seq->patterns; i++)
	seq->pattern[i] = PS04_pattern_new();
    for (i=0; i < seq->songs; i++)
	seq->song[i] = PS04_song_new();

    return seq;
}



ps04_seq_t *PS04_seq_open(const char *file)
{
    ps04_seq_t *newseq;

    newseq = malloc(sizeof(*newseq));
    memset(newseq, 0, sizeof(*newseq));

    newseq->F = fopen(file, "r+b");
    if (newseq->F == NULL)
    {
	free(newseq);
	return NULL;
    }

    PS04_seq_read(newseq);
    return newseq;
}



void PS04_seq_close(ps04_seq_t *seq)
{
    fclose(seq->F);
    free(seq);
}



/* Conversion routines */

static void PS02_pattern_2_PS04_pattern(ps02_pattern_t *ptn02, ps04_pattern_t *ptn04)
{
    memset(ptn04->name_long, 0x20, sizeof(ptn04->name_long));
    memcpy(ptn04->name_long, ptn02->name, 6);
    memset(ptn04->name, 0x20, sizeof(ptn04->name));
    memcpy(ptn04->name, ptn02->name, 6);
    ptn04->beats = ptn02->beats;
    ptn04->bars = ptn02->bars;
    ptn04->drum_kit = 0;
    ptn04->bass_kit = 0;
    ptn04->drum_level = 0x0f;
    ptn04->bass_level = 0x0f;
    ptn04->original_root = 9;	/* A */
    ptn04->original_chord = 0;
    ptn04->event_start = 0;
    ptn04->rawevents = ptn02->events;
    ptn04->event = ptn02->event;
}



static int PS02_volume_2_PS04_volume(int ps02)
{
    /* PS02 : 0-63
     * PS04 : 0-15
     */
    return ps02 >> 2;
}



static void PS02_song_2_PS04_song(ps02_song_t *song02, ps04_song_t *song04,
				  ps02_pts_t *pts)
{
    int			 measure;
    song_event_t	*event;
    int			 ppq_delta;
    unsigned short	*ppq;
    int			 timesig = -1;

    memset(song04->name, 0, sizeof(song04->name));
    memcpy(song04->name, song02->name, sizeof(song02->name));
    memset(song04->name_long, 0, sizeof(song04->name_long));
    memcpy(song04->name_long, song02->name, sizeof(song02->name));

    song04->event = realloc(song04->event, (song02->measures *
 			   MAX_BEATS_PER_MEASURE * 2 + 3) *
 			   sizeof(*song04->event));
    event = song04->event;

    ppq_delta = 0;
    ppq = &event->next_ppq;

    for(measure = 0; measure < song02->measures; measure++)
    {
	int beat;

	if (song02->measure[measure].pattern != 0xff)
	{
	    if (measure != 0)
	    {
		(event-1)->next_ppq = ppq_delta;
		ppq_delta = 0;
	    }
	    event->type = SONG_EVENT_PATTERN;
	    event->data = song02->measure[measure].pattern;
	    event->next_ppq = 0;
	    event++;
	}

	if (measure == 0)
	{
	    event->type = SONG_EVENT_TEMPO;
	    event->data = song02->tempo * 10;
	    event->next_ppq = 0;
	    event++;

	    event->type = SONG_EVENT_DRUM_VOL;
	    event->data = PS02_volume_2_PS04_volume(song02->drum_volume);
	    event->next_ppq = 0;
	    event++;

	    event->type = SONG_EVENT_BASS_VOL;
	    event->data = PS02_volume_2_PS04_volume(song02->bass_volume);
	    event->next_ppq = 0;
	    event++;
	}

	if (timesig != pts->rhy_pattern[song02->measure[0].pattern]->beats)
	{
	    timesig = pts->rhy_pattern[song02->measure[0].pattern]->beats;
	    event->type = SONG_EVENT_TIMESIG;
	    event->data = timesig;
	    event->next_ppq = 0;
	    event++;
	}

	for(beat = 0; beat < song02->measure[measure].beats; beat++)
	{
	    if (song02->measure[measure].note[beat] != 0xff)
	    {
		(event-1)->next_ppq = ppq_delta;
		ppq_delta = 0;
		event->type = SONG_EVENT_NOTE;
		event->data = song02->measure[measure].note[beat];
		ppq = &event->next_ppq;
		event++;
	    }
	    ppq_delta += PS04_PPQ;
	}
    }
    (event-1)->next_ppq = ppq_delta;
    song04->events = event - song04->event;
}



void PS02_to_PS04(ps02_pts_t *pts, ps04_seq_t *seq, int songindx)
{
    int i;

    for(i =  0; i < pts->rhy_pattern_cnt; i++)
	PS02_pattern_2_PS04_pattern(pts->rhy_pattern[i], seq->pattern[i]);

    for(; i < seq->patterns; i++)
    {
	ps04_pattern_t *ptn = seq->pattern[i];
	if (ptn->raw)
	    free(ptn->raw);
	memset(ptn, 0, sizeof(*ptn));
	strcpy(ptn->name_long, "(empty)");
    }

    PS02_song_2_PS04_song(pts->song[songindx], seq->song[0], pts);
}

void MRS8_bin_read(mrs8_bin_t *bin)
{
    FILE	*in;
    int		 i;
    uint32_t	 size;

    in = bin->F;

    GET_FIELD(bin->unknown1);
    GET_FIELD(bin->drum_kits);
    GET_FIELD(bin->bass_kits);
    GET_FIELD(bin->samples);

    Log(2, "Samples:%d\n", bin->samples);

    bin->sample_def = malloc(bin->samples * sizeof(*bin->sample_def));
    fread(bin->sample_def, bin->samples * sizeof(*bin->sample_def), 1, in);

    /* Determine total number of samples */
    bin->total_frames = 0;
    for(i = 0; i < bin->samples; i++)
    {
	mrs8_sample_def_t *sd = &bin->sample_def[i];
	bin->total_frames += sd->frames;
    }

    /* Even though we know the real size, malloc enough to hold
     * the max possible samples in case they want to import more
     * than what it currently has.
     */
    size = bin->total_frames * sizeof(*bin->sample_data);
    bin->sample_data = malloc(4*1024*1024);
    fread(bin->sample_data, size, 1, in);
}

void MRS8_bin_write(mrs8_bin_t *bin)
{
    FILE	*out;
    long	 pos;
    long	 size;

    out = bin->F;
    fseek(out, 0, SEEK_SET);
    Ftruncate(fileno(out), 0);

    PUT_FIELD(bin->unknown1);
    PUT_FIELD(bin->drum_kits);
    PUT_FIELD(bin->bass_kits);
    PUT_FIELD(bin->samples);
    fwrite(bin->sample_def, bin->samples * sizeof(*bin->sample_def), 1, out);
    fwrite(bin->sample_data, bin->total_frames * sizeof(*bin->sample_data), 1, out);

    /* Force file size mod 4MB */
    pos = ftell(out);
    size = (pos + 0x3fffff) & ~0x3fffff;
    PUT_ZEROS(size-pos);
}

mrs8_bin_t *MRS8_bin_open(const char *file)
{
    mrs8_bin_t *newbin;

    newbin = malloc(sizeof(*newbin));
    memset(newbin, 0, sizeof(*newbin));

    newbin->F = fopen(file, "r+b");
    if (newbin->F == NULL)
    {
	free(newbin);
	return NULL;
    }

    MRS8_bin_read(newbin);
    return newbin;
}

void MRS8_bin_close(mrs8_bin_t *bin)
{
    fclose(bin->F);
    free(bin->sample_data);
    free(bin->sample_def);
    free(bin);
}




/* .AUD or .ZAF -> .WAV conversion utilities */

/* Both .aud and .zaf files use 16 bit samples */
#define SAMPLE_BITS 16
static int deltas[] = {
  0,1,2,5,9,14,19,0,24,29,
  34,40,45,51,57,62,68,75,81,87,
  94,100,107,114,120,127,134,141,148,155,
  163,170,177,185,192,200,208,215,223,231,
  239,246,254,262,271,279,288,297,306,316,
  325,336,346,357,368,380,391,404,416,429,
  443,457,471,486,501,516,533,549,566,584,
  602,621,640,660,681,702,724,747,770,794,
  819,845,871,898,926,955,985,1016,1048,1081,
  1115,1151,1188,1226,1266,1308,1351,1397,1444,1493,
  1544,1598,1653,1712,1773,1836,1903,1972,2045,2121,
  2204,2297,2402,2520,2654,2810,2991,3207,3468,3792,
  4206,4732,5324,5989,6738,7580,8527,9551,-9551,-8527,
  -7580,-6738,-5989,-5324,-4732,-4206,-3792,-3468,-3207,-2991,
  -2810,-2654,-2520,-2402,-2297,-2204,-2121,-2045,-1972,-1903,
  -1836,-1773,-1712,-1653,-1598,-1544,-1493,-1444,-1397,-1351,
  -1308,-1266,-1226,-1188,-1151,-1115,-1081,-1048,-1016,-985,
  -955,-926,-898,-871,-845,-819,-794,-770,-747,-724,
  -702,-681,-660,-640,-621,-602,-584,-566,-549,-533,
  -516,-501,-486,-471,-457,-443,-429,-416,-404,-391,
  -380,-368,-357,-346,-336,-325,-316,-306,-297,-288,
  -279,-271,-262,-254,-246,-239,-231,-223,-215,-208,
  -200,-192,-185,-177,-170,-163,-155,-148,-141,-134,
  -127,-120,-114,-107,-100,-94,-87,-81,-75,-68,
  -62,-57,-51,-45,-40,-34,-29,-24,0,-19,
  -14,-9,-5,-2,-1,0
};

void
write_wave_header(FILE *out, size_t frames, int channels, int sample_rate,
		  int sample_bits)
{
    WaveHeader	header;
    DataChunk	data;
    uint32_t	bytes;

    bytes = frames * channels * (sample_bits / 8);
    memset(&header, 0, sizeof(header));
    memcpy(header.group_id,  "RIFF", 4);
    memcpy(header.riff_type, "WAVE", 4);
    header.size = bytes+sizeof(header);

    memcpy(header.fmt.chunk_id, "fmt ", 4);
    header.fmt.chunk_size = sizeof(FmtChunk) - 8;
    header.fmt.format = WAVE_FORMAT_PCM;
    header.fmt.channels = channels;
    header.fmt.sample_rate = sample_rate;
    header.fmt.block_align = channels * (sample_bits / 8);
    header.fmt.avg_bytes_per_second = sample_rate * header.fmt.block_align;
    header.fmt.bits_per_sample = sample_bits;

    memcpy(data.chunk_id, "data", 4);
    data.chunk_size = bytes;

    fwrite(&header, 1, sizeof(header), out);
    fwrite(&data, 1, sizeof(data), out);
}

int read_wave_header(FILE *in, WaveHeader *header, DataChunk *data)
{
    memset(header, 0, sizeof(*header));
    fread(header, 1, sizeof(*header), in);

    if (memcmp(header->group_id,  "RIFF", 4) ||
	memcmp(header->riff_type, "WAVE", 4) ||
	memcmp(header->fmt.chunk_id, "fmt ", 4))
    {
	printf("Unable to identify wav file\n");
	return FALSE;
    }

    memset(data, 0, sizeof(*data));
    fread(data, 1, sizeof(*data), in);
    if (memcmp(data->chunk_id, "data", 4))
    {
	printf("Unable to find data chunk\n");
	return FALSE;
    }
    return TRUE;
}

static int
safe_fgetc(FILE *f)
{
    int rc = fgetc(f);
    if (feof(f))
    {
	fprintf(stderr, "ERROR: Unexpected early EOF!\n");
	exit(-1);
    }
    return rc;
}

static void
audblk2samples(FILE *f, short *sample_buf, int zaf_mode)
{
    short sample = 0;
    int bufIndex = 0;
    int hibyte;
    int lobyte;

    int i;
    int deltaIndex;
    int delta;

    /* Fill smp_buf with 509 samples from AUD data */

    /* Read the absolute sample value from the AUD */
    lobyte = fgetc(f);
    if (lobyte == EOF)
    {
	static int warn = 0;
	if (!warn)
	{
	    warn = 1;
	    fprintf(stderr, "WARNING: Reached early EOF padding with zero (aud files not same size?)\n");
	}
	return;
    }
    hibyte = safe_fgetc(f);
    if (zaf_mode)
	sample = (short)((hibyte<<8) | (lobyte & 0xff));
    else
	sample = (short)((lobyte<<8) | (hibyte & 0xff));

    sample_buf[bufIndex++] = sample;

    /* Calculate the next 508 samples from the deltas */
    for (i = 0; i < 508; i++)
    {
	deltaIndex = safe_fgetc(f);
	delta = deltas[deltaIndex];
	sample += delta;
	sample_buf[bufIndex++] = sample;
    }

    /* Skip the terminator bytes */
    safe_fgetc(f);
    safe_fgetc(f);
}

void auds2wav(FILE *in[2], FILE *out, int channels, void (*cb)(int percent))
{
    uint32_t	blocks;
    uint32_t	frames;
    uint32_t	sample_rate;
    uint32_t	length;
    size_t	samples_size = 0;
    long	header_size;
    int		percent_done = -1;
    int		i,j,k;
    int		zaf_mode;
    char	magic[13];
    
    /* Sample the first bit to determine if its a .zaf file */
    fread(&magic, 1, sizeof(magic), in[0]);
    if (!strncmp(magic, "ZOOM1D5ZZ3AUD", sizeof(magic)))
    {
	/* ZAF file */
	zaf_mode = TRUE;
	sample_rate = 44100;
	header_size = 16384;
    }
    else
    {
	/* AUD file */
	zaf_mode = FALSE;
	sample_rate = 31250;
	header_size = 512;
    }

    for(i=0; i < channels; i++)
    {
	struct stat sbuf;

	/* Skip the header */
	fseek(in[i], header_size, SEEK_SET);

	/* Get max length of any input file */
	fstat(fileno(in[i]), &sbuf);
	samples_size = MAX(samples_size, sbuf.st_size);
    }

    length  = samples_size - header_size;
    blocks  = length / 512;
    frames  = blocks * 509;

    /* Create .wav header */
    write_wave_header(out, frames, channels, sample_rate, SAMPLE_BITS);

    /* Process all 512-byte data blocks from the file until EOF found */
    for(i = 0; i < blocks; i++)
    {
	short samples[2][509];

	memset(samples, 0, sizeof(samples));
	if (percent_done != (i * 100) / blocks)
	{
	    percent_done = (i * 100) / blocks;
	    if (cb)
		cb(percent_done);
	}
	for(j = 0; j < channels; j++)
	    audblk2samples(in[j], &samples[j][0], zaf_mode);

	for(j = 0; j < 509; j++)
	{
	    for(k = 0; k < channels; k++)
	    {
		fputc(samples[k][j]&0xff, out); 
		fputc((samples[k][j]>>8)&0xff, out); 
	    }
	}
    }
}



#ifdef MINGW
#include <errno.h>
#include <windows.h>

static int Ftruncate(int fd, int size)
{
  HANDLE hfile;
  unsigned int curpos;
                                                                                               
  hfile = (HANDLE) _get_osfhandle (fd);
  curpos = SetFilePointer (hfile, 0, NULL, FILE_CURRENT);
  if (curpos == 0xFFFFFFFF
      || SetFilePointer (hfile, size, NULL, FILE_BEGIN) == 0xFFFFFFFF
      || !SetEndOfFile (hfile))
    {
      int error = GetLastError ();
                                                                                               
      switch (error)
        {
        case ERROR_INVALID_HANDLE:
          errno = EBADF;
          break;
        default:
          errno = EIO;
          break;
        }
                                                                                               
      return -1;
    }
}
#else
#include <unistd.h>
static int Ftruncate(int fd, off_t length)
{
    return ftruncate(fd, length);
}
#endif
