/*
 * libzoom - Manage the data of a Zoom PS-02 or PS-04.
 *
 * Copyright (C) 2003-2004 Dwight Engen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: $
 */

/*
 * General Notes
 * =============
 *   Event resolution	PPQ (Parts Per Quarter note)
 *   -----------------	----------------------------
 *   PS02		24
 *   PS04		48
 *   ps0xlib internal	48
 *
 * MIDI note numbers
 * =================
 *   Octave||                     Note Numbers
 *      #  ||
 *         || C   | C#  | D   | D#  | E   | F   | F#  | G   | G#  | A   | A#  | B
 *   ------++-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+-----+----
 *     -1  ||   0 |   1 |   2 |   3 |   4 |   5 |   6 |   7 |   8 |   9 |  10 |  11
 *      0  ||  12 |  13 |  14 |  15 |  16 |  17 |  18 |  19 |  20 |  21 |  22 |  23
 *      1  ||  24 |  25 |  26 |  27 |  28 |  29 |  30 |  31 |  32 |  33 |  34 |  35
 *      2  ||  36 |  37 |  38 |  39 |  40 |  41 |  42 |  43 |  44 |  45 |  46 |  47
 *      3  ||  48 |  49 |  50 |  51 |  52 |  53 |  54 |  55 |  56 |  57 |  58 |  59
 *      4  ||  60 |  61 |  62 |  63 |  64 |  65 |  66 |  67 |  68 |  69 |  70 |  71
 *      5  ||  72 |  73 |  74 |  75 |  76 |  77 |  78 |  79 |  80 |  81 |  82 |  83
 *      6  ||  84 |  85 |  86 |  87 |  88 |  89 |  90 |  91 |  92 |  93 |  94 |  95
 *      7  ||  96 |  97 |  98 |  99 | 100 | 101 | 102 | 103 | 104 | 105 | 106 | 107
 *      8  || 108 | 109 | 110 | 111 | 112 | 113 | 114 | 115 | 116 | 117 | 118 | 119
 *      9  || 120 | 121 | 122 | 123 | 124 | 125 | 126 | 127 |
 *
 * PS04 MIDI drum note numbers
 * ===========================
 *   Note Name             Note Name
 *   ---- ------------     ---- -------------
 *   31   Clave            50   Tom1 high
 *   33   Met click        51   Ride cymbal
 *   34   Met bell         53   Ride bell
 *   36   Kick drum 1      54   Tambourine
 *   37   Side stick       56   Cow bell
 *   38   Snare 1          60   Bongo high
 *   39   Handclap         61   Bongo low
 *   41   Tom2 low         62   Conga slap
 *   42   HiHat closed     63   Conga hi
 *   43   Tom1 low         64   Conga low
 *   45   Tom2 mid         65   Timbale high
 *   46   HiHat open       66   Timbale low
 *   47   Tom1 mid         67   Agogo high
 *   48   Tom2 high        70   Maracas
 *   49   Crash            73   Guiro (short)
 *
 * PS04 MIDI bass note numbers
 * ===========================
 * 0xc0 C1?
 * 0xd0 E1
 * 0xe0 G#2
 */


/*
 * PS02 bass bank and note to MIDI note mapping
 * ============================================
 *        PS02             MIDI     0x80 MIDI
 *   Sounds  Bank:Note  Note  Name  Note  Name
 *   ------  ---------  ----  ----  ----  ----
 *   No      0x0e:0x0   21    A0    22    A#0
 *   No      0x0e:0x1   23    B0    23    B0
 *   Yes     0x0e:0x2   24    C1    26    D1
 *   Yes     0x0e:0x3   27    D#1   28    E1
 *   Yes     0x0e:0x4   29    F1    31    G1
 *   Yes     0x0e:0x5   33    A1    32    G#1
 *   Yes     0x0e:0x6   33    A1    34    A#1
 *   Yes     0x0e:0x7   35    B1    35    B1
 *   Yes     0x0e:0x8   36    C2    38    D2
 *   Yes     0x0e:0x9   39    D#2   28    E1
 *   Yes     0x0e:0xa   29    F1    31    G1
 *   Yes     0x0e:0xb   33    A1    32    G#1
 *   Yes     0x0e:0xc   33    A1    34    A#1
 *   Yes     0x0e:0xd   35    B1    35    B1
 *
 *   Yes     0x1e:0x0   33    A1
 *   Yes     0x1e:0x1   35    B1
 *   Yes     0x1e:0x2   36    C2
 *   Yes     0x1e:0x3   39    D#2
 *   Yes     0x1e:0x4   41    F2
 *   Yes     0x1e:0x5   45    A2
 *   Yes     0x1e:0x6   45    A2
 *   Yes     0x1e:0x7   35    B1
 *   Yes     0x1e:0x8   36    C2
 *   Yes     0x1e:0x9   39    D#2
 *   Yes     0x1e:0xa   41    F2
 *   Yes     0x1e:0xb   45    A2
 *   Yes     0x1e:0xc   45    A2
 *   Yes     0x1e:0xd   47    B2
 *
 *   Yes     0x2e:0x0   33    A1
 *   Yes     0x2e:0x1   35    B1
 *   Yes     0x2e:0x2   36    C2
 *   Yes     0x2e:0x3   39    D#2
 *   Yes     0x2e:0x4   41    F2
 *   Yes     0x2e:0x5   45    A2
 *   Yes     0x2e:0x6   45    A2
 *   Yes     0x2e:0x7   47    B2
 *   Yes     0x2e:0x8   48    C3
 *   Yes     0x2e:0x9   51    D#3
 *   Yes     0x2e:0xa   41    F2
 *   Yes     0x2e:0xb   45    A2
 *   Yes     0x2e:0xc   45    A2
 *   Yes     0x2e:0xd   47    B2
 */

/*
 * PS02 drum bank and note to MIDI note mapping
 * ============================================
 *        PS02             MIDI
 *   Sample  Bank:Note  Note  Name  Comments
 *   ------  ---------  ----  ----  --------
 *   No      0x0e:0x0   24
 *   No      0x0e:0x1   26
 *   No      0x0e:0x2   28
 *   No      0x0e:0x3   30
 *   Yes     0x0e:0x4   32          Not MIDI, Bass slide
 *   Yes     0x0e:0x5   34
 *   Yes     0x0e:0x6   36
 *   Yes     0x0e:0x7   38
 *   Yes     0x0e:0x8   40
 *   Yes     0x0e:0x9   42
 *   Yes     0x0e:0xa   44
 *   Yes     0x0e:0xb   46
 *   No      0x0e:0xc   48
 *   Yes     0x0e:0xd   50
 *
 *   Yes     0x1e:0x0   36
 *   Yes     0x1e:0x1   38
 *   Yes     0x1e:0x2   40
 *   Yes     0x1e:0x3   42
 *   Yes     0x1e:0x4   44
 *   Yes     0x1e:0x5   46
 *   No      0x1e:0x6   48
 *   Yes     0x1e:0x7   50
 *   No      0x1e:0x8   52
 *   No      0x1e:0x9   54
 *   Yes     0x1e:0xa   56
 *   No      0x1e:0xb   58
 *   Yes     0x1e:0xc   60
 *   Yes     0x1e:0xd   62
 *
 *   No      0x2e:0x0   48
 *   Yes     0x2e:0x1   50
 *   No      0x2e:0x2   52
 *   Yes     0x2e:0x3   54
 *   Yes     0x2e:0x4   56
 *   No      0x2e:0x5   58
 *   Yes     0x2e:0x6   60
 *   Yes     0x2e:0x7   62
 *   No      0x2e:0x8   64
 *   Yes     0x2e:0x9   66
 *   Yes     0x2e:0xa   68
 *   No      0x2e:0xb   70
 *   No      0x2e:0xc   72
 *   No      0x2e:0xd   74
 *
 *   No      0x3e:0x0   60
 *   Yes     0x3e:0x1   62
 *   No      0x3e:0x2   64
 *   Yes     0x3e:0x3   66
 *   Yes     0x3e:0x4   68
 *   No      0x3e:0x5   70
 *   Yes     0x3e:0x6   72
 *   Yes     0x3e:0x7   74
 *   No      0x3e:0x8   76
 *   Yes     0x3e:0x9   78
 *   Yes     0x3e:0xa   80
 *   No      0x3e:0xb   82
 *   No      0x3e:0xc   84
 *   No      0x3e:0xd   86
 */


#ifndef _LIBZOOM_H

#include <stdio.h>
#include <stdint.h>

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#define FALSE 0
#define TRUE 1

#define PACKED __attribute__((packed))

#define MAX_BEATS_PER_MEASURE	5
#define MAX_MEASURES		1024
#define MAX_EVENTS_PER_PATTERN	255
#define NUM_SONGS		100
#define NUM_PATCHES		60

#define PS02_PPQ_FACTOR		2
#define PS04_PPQ_FACTOR		1
#define PS02_PPQ		24
#define PS04_PPQ		48

#define Log(level, fmt, ...)					\
do {								\
    if (level <= DEBUG)						\
    {								\
	fprintf(stdout, "%-24.24s:" fmt,			\
		__PRETTY_FUNCTION__, ## __VA_ARGS__);		\
    }								\
} while(0)

#define Assert(assert, fmt, ...)				\
    (void)((assert) ? 0 :					\
	fprintf(stderr, "Assert @ %s:%d:%s(): " fmt,		\
		__FILE__, __LINE__,				\
		__PRETTY_FUNCTION__, ## __VA_ARGS__))


/* Generic software structures */

typedef enum {
    PTN_EVENT_DRUM,
    PTN_EVENT_DRUM_BANK,
    PTN_EVENT_BASS,
    PTN_EVENT_BASS_BANK,
    PTN_EVENT_PPQ,
    PTN_EVENT_END,
} ptn_event_type_t;

typedef struct ptn_event_s
{
    ptn_event_type_t	 type;
    unsigned int	 ppq_delta;
    unsigned int	 ppq_duration;
    unsigned int	 midi_note;
    unsigned int	 velocity15;
} ptn_event_t;

typedef enum {
    /* PS04 hardware values */
    SONG_EVENT_PATTERN =	   1,
    SONG_EVENT_NOTE,		/* 2 */
    SONG_EVENT_TEMPO,		/* 3 */
    SONG_EVENT_TIMESIG,		/* 4 */
    SONG_EVENT_DRUM_VOL,	/* 5 */
    SONG_EVENT_BASS_VOL,	/* 6 */
    SONG_EVENT_DRUM_KIT,	/* 7 */
    SONG_EVENT_BASS_KIT,	/* 8 */
} song_event_type_t;

typedef struct song_event_s
{
    song_event_type_t	 type;
    unsigned short	 data;
    unsigned short	 next_ppq;
} song_event_t;


/* PS02 specific structures */

typedef struct ps02_measure_s
{
    unsigned char	pattern;
    unsigned char	beats;
    unsigned char	note[MAX_BEATS_PER_MEASURE];
} ps02_measure_t;

typedef struct ps02_pattern_s
{
    /* Hardware defined structure */
    char		 name[6];
    char		 name_pad[250];
    unsigned char	 bpm;
    unsigned char	 beats;
    unsigned char	 bars;
    unsigned char	 unknown2[6];
    unsigned char	 bass_octave;
    unsigned char	 unknown3[3];
    unsigned long	 raw_bytes;
    unsigned char	*raw;

    /* Unpacked pattern events */
    unsigned int	 events;
    ptn_event_t		*event;
} ps02_pattern_t;

typedef struct ps02_song_s
{
    /* Hardware defined structure */
    char		 name[6];
    unsigned char	 drum_volume;
    unsigned char	 bass_volume;
    unsigned char	 input_volume;
    unsigned char	 pad1[247];
    unsigned long	 patterns_len;
    unsigned char	 tempo;
    unsigned char	 repeat;
    unsigned char	 drum_kit;
    unsigned char	 bass_kit;
    unsigned char	*raw;

    /* Software defined structure */
    unsigned char	 measures;
    ps02_measure_t	*measure;
} PACKED ps02_song_t;

typedef struct ps02_pts_s
{
    char		 header[0x250];
    unsigned long	 rhy_pattern_cnt;
    ps02_pattern_t	*rhy_pattern[0x100];
    char		 rhy_grp0_indx[0x100];
    char		 rhy_grp1_indx[0x100];
    char		 rhy_grp2_indx[0x100];
    char		 rhy_main_group_cnt;
    char		*rhy_main_group_names;
    char		 rhy_sub_group_cnt;
    char		*rhy_sub_group_names;
    ps02_song_t		*song[0x100];
    int			 modified;
    FILE		*F;
} PACKED ps02_pts_t;

typedef struct patch_s
{
    char		 name[7];
    char		 data[18];
} PACKED patch_t;

typedef struct pch_s
{
    patch_t		 pch[NUM_PATCHES];
    int			 modified;
    FILE		*F;
} ps02_pch_t;



/* PS04 structures */


typedef struct ps04_pattern_s
{
    /* Hardware defined structure */
    char		 name_long[0x100];
    char		 name[8];
    unsigned char	 beats;
    unsigned char	 bars;
    unsigned char	 drum_kit;
    unsigned char	 bass_kit;
    unsigned char	 drum_level;
    unsigned char	 bass_level;
    unsigned char	 original_root;
    unsigned char	 original_chord;
    unsigned short	 rawevents;
    unsigned short	 event_start;
    unsigned char	*raw;

    /* Software defined fields */
    /* Unpacked pattern events */
    unsigned short	 sevents;
    ptn_event_t		*event;

    unsigned int	 starting_ppq;
} PACKED ps04_pattern_t;

typedef struct ps04_song_event_s
{
    uint8_t		 type;
    unsigned short	 data;
    uint8_t		 next_ppq;
} PACKED ps04_song_event_t;

typedef struct ps04_song_s
{
    /* Hardware defined structure */
    char		 name_long[0x100];
    char		 name[8];
    unsigned short	 rawevents;
    unsigned char	 unknown1[2];
    unsigned char	*raw;

    /* Software defined fields */
    unsigned short	 events;
    song_event_t	*event;
} PACKED ps04_song_t;

typedef struct ps04_seq_s
{
    /* Hardware defined structure */
    char		 fmt[13];	/* "ZOOM D120SEQ\000" */
    unsigned char	 unknown1[7];
    unsigned long	 patterns;	/* 511 */
    unsigned long	 songs;		/* 10 */
    unsigned char	 unknown2[0x1e4];
    ps04_pattern_t	*pattern[511];
    ps04_song_t		*song[10];

    /* Software defined fields */
    int			 modified;
    FILE		*F;
} PACKED ps04_seq_t;

/* MRS8 structures */

#define MRS8_DRUM_KITS		32
#define MRS8_DRUMS_PER_KIT	36
#define MRS8_DRUM_KITS_OFFSET	0x127ec6
#define MAX_SAMPLES		0x3e

typedef struct mrs8_drum_s
{
    uint8_t		sample;
    uint8_t		midi_note;
    int8_t		pan;
    uint8_t		unknown1;	/* Volume ? */
    int32_t		pitch;
} PACKED mrs8_drum_t;

typedef struct mrs8_drum_kit_s
{
    char		name[8];
    mrs8_drum_t		drum[MRS8_DRUMS_PER_KIT];
} PACKED mrs8_drum_kit_t;

typedef struct mrs8_drum_kits_s
{
    mrs8_drum_kit_t	kit[MRS8_DRUM_KITS];
} PACKED mrs8_drum_kits_t;

typedef struct mrs8_bass_kit_s
{
    char		name[8];
    char		unknown[0x80];
} PACKED mrs8_bass_kit_t;

typedef struct mrs8_bass_kits_s
{
    mrs8_bass_kit_t	kit[32];
} PACKED mrs8_bass_kits_t;

typedef struct mrs8_sample_def_s
{
    uint32_t		frames;
    uint32_t		frame_loop_begin;
    uint32_t		frame_loop_end;
    uint32_t		unknown1;
} PACKED mrs8_sample_def_t;

typedef struct mrs8_bin_s
{
    /* Hardware defined structure */
    char		 unknown1[0x127ec6];
    mrs8_drum_kits_t	 drum_kits;
    mrs8_bass_kits_t	 bass_kits;
    uint32_t		 samples;
    mrs8_sample_def_t	*sample_def;
    short		*sample_data;

    /* Software defined fields */
    FILE		*F;
    int			 modified;
    uint32_t		 total_frames;
} PACKED mrs8_bin_t;

/* IFF .wav format structures */

#define WAVE_FORMAT_PCM 1
typedef struct _DataChunk	DataChunk;
typedef struct _FmtChunk	FmtChunk;
typedef struct _WaveHeader	WaveHeader;

struct _FmtChunk
{
    char	chunk_id[4];		/* "fmt " */
    uint32_t 	chunk_size;
    uint16_t	format;
    uint16_t	channels;
    uint32_t	sample_rate;
    uint32_t	avg_bytes_per_second;
    uint16_t	block_align;
    uint16_t	bits_per_sample;    
};

struct _DataChunk
{
    char	chunk_id[4];		/* "data" */
    uint32_t	chunk_size;
};

struct _WaveHeader
{
    char	group_id[4];		/* "RIFF" */
    uint32_t	size;
    char	riff_type[4];		/* "WAVE" */
    FmtChunk	fmt;
};

extern const char note_root_name[][4];
extern const char note_tone_name[][4];

ps02_pch_t	*PS02_pch_open(const char *file);
void		 PS02_pch_close(ps02_pch_t *pch);
void		 PS02_pch_read(FILE *in, ps02_pch_t *pch);
void		 PS02_pch_write(FILE *out, ps02_pch_t *pch);
void		 PS02_patch_read(FILE *in, patch_t *patch);
void		 PS02_patch_write(FILE *out, patch_t *patch);
ps02_pts_t	*PS02_pts_open(const char *file);
void		 PS02_pts_close(ps02_pts_t *pts);
int		 PS02_pts_modified(ps02_pts_t *pts);
void		 PS02_pts_write(ps02_pts_t *pts);
void		 PS02_pts_read(ps02_pts_t *pts);
void		 PS02_song_copy_measures(ps02_song_t *song, int start, int end, int dest);
void		 PS02_song_delete_measures(ps02_song_t *song, int start, int end);
void		 PS02_song_double(ps02_song_t *song);
void		 PS02_song_free(ps02_song_t *song);
void		 PS02_song_write(FILE *out, ps02_song_t *song);
ps02_song_t	*PS02_song_copy(ps02_song_t *origsong);
ps02_song_t	*PS02_song_read(FILE *in);

void		 PS04_pattern_write(FILE *out, ps04_pattern_t *ptn);
ps04_pattern_t	*PS04_pattern_read(FILE *in);
ps04_pattern_t	*PS04_pattern_new(void);
void		 PS04_song_write(FILE *out, ps04_song_t *song);
ps04_song_t	*PS04_song_read(FILE *in);
ps04_song_t	*PS04_song_new(void);
void		 PS04_seq_write(ps04_seq_t *seq);
void		 PS04_seq_read(ps04_seq_t *seq);
ps04_seq_t	*PS04_seq_new(void);
ps04_seq_t	*PS04_seq_open(const char *file);
void		 PS04_seq_close(ps04_seq_t *seq);

mrs8_bin_t	*MRS8_bin_open(const char *file);
void		 MRS8_bin_close(mrs8_bin_t *bin);
void		 MRS8_bin_read(mrs8_bin_t *bin);
void		 MRS8_bin_write(mrs8_bin_t *bin);


void		 PS02_to_PS04(ps02_pts_t *pts, ps04_seq_t *seq, int songindx);

const char	*PS0x_midi_drum_name(int midi_note);
const char	*PS0x_midi_note_name(int midi_note);

void		 auds2wav(FILE *in[2], FILE *out, int channels, void (*cb)(int percent));
int		 read_wave_header(FILE *in, WaveHeader *header, DataChunk *data);
void		 write_wave_header(FILE *out, size_t frames, int channels,
				   int sample_rate, int sample_bits);
void		 hexdump(const void *inbuf, size_t bytes);

#endif /* _LIBZOOM_H */
